#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dialogs.export.py
"""
GUI tools to export text data to a file
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
import wx.adv
from functools import partial
import matplotlib.colors as plot_colors
from pubsub import pub
from dvha.db.sql_connector import DVH_SQL
from dvha.models.data_table import DataTable
from dvha.paths import DATA_DIR
from dvha.tools.threading_progress import ProgressFrame
from dvha.tools.utilities import (
    get_selected_listctrl_items,
    save_object_to_file,
    set_msw_background_color,
    set_frame_icon,
)


def save_data_to_file(
    frame,
    title,
    data,
    wildcard="CSV files (*.csv)|*.csv",
    data_type="string",
    initial_dir=DATA_DIR,
):
    """
    from https://wxpython.org/Phoenix/docs/html/wx.FileDialog.html
    :param frame: GUI parent
    :param title: title for the file dialog window
    :type title: str
    :param data: text data or pickle-able object to be written
    :param wildcard: restrict visible files and intended file extension
    :type wildcard: str
    :param data_type: either 'string' or 'pickle'
    :type data_type: str
    :param initial_dir: start the FileDialog at this directory
    :type initial_dir: str
    """

    with wx.FileDialog(
        frame,
        title,
        defaultDir=initial_dir,
        wildcard=wildcard,
        style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT,
    ) as fileDialog:

        if fileDialog.ShowModal() == wx.ID_CANCEL:
            return

        pathname = fileDialog.GetPath()

        if data_type == "string":
            try:
                with open(pathname, "w", encoding="utf-8") as file:
                    file.write(data)
            except IOError:
                wx.LogError(
                    "Cannot save current data in file '%s'." % pathname
                )

        elif data_type == "pickle":
            save_object_to_file(data, pathname)

        elif data_type == "function":
            data(pathname)

        return pathname


class ExportCSVDialog(wx.Dialog):
    """
    Allow user to select available data for export to CSV into one file.
    This class leverages the get_csv functions built-into each of the data types / tabs
    """

    def __init__(self, app):
        """
        :param app: easier to pass main frame pointer than several links to each data type
        :type app: DVHAMainFrame
        """
        wx.Dialog.__init__(self, None)

        self.app = app

        # Each of these objects shoudl have a has_data property, if False, those UI elements in this class will
        # be disabled (i.e., don't allow user to export empty tables
        self.enabled = {
            "DVHs": self.app.group_data[1]["dvh"].has_data,
            "DVHs Summary": self.app.group_data[1]["dvh"].has_data,
            "Endpoints": self.app.endpoint.has_data,
            "Radbio": self.app.radbio.has_data,
            "Charting Variables": self.app.time_series.has_data,
        }

        checkbox_keys = [
            "DVHs",
            "DVHs Summary",
            "Endpoints",
            "Radbio",
            "Charting Variables",
        ]
        self.checkbox = {
            key: wx.CheckBox(self, wx.ID_ANY, key) for key in checkbox_keys
        }

        # set to a dictionary because previous versions had a tree with Regression
        self.list_ctrl = {
            "Charting Variables": wx.ListCtrl(
                self,
                wx.ID_ANY,
                style=wx.LC_HRULES | wx.LC_REPORT | wx.LC_VRULES,
            )
        }

        time_series_column = "Variables"
        time_series_variables = (
            self.app.time_series.combo_box_y_axis.GetItems()
        )
        time_series_data = {time_series_column: time_series_variables}
        self.data_table_time_series = DataTable(
            self.list_ctrl["Charting Variables"],
            columns=[time_series_column],
            widths=[400],
        )
        self.data_table_time_series.set_data(
            time_series_data, [time_series_column]
        )

        # set to a dictionary because previous versions had a table with Regression
        self.button_select_data = {
            "Charting Variables": {
                "Select": wx.Button(self, wx.ID_ANY, "Select All"),
                "Deselect": wx.Button(self, wx.ID_ANY, "Deselect All"),
            }
        }

        self.button_save = wx.Button(self, wx.ID_OK, "Save")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")
        self.button_select_all = wx.Button(self, wx.ID_ANY, "Select All")
        self.button_deselect_all = wx.Button(self, wx.ID_ANY, "Deselect All")

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetTitle("Export Data to CSV")
        self.button_select_all.SetToolTip(
            "Only data objects with data will be enabled."
        )
        self.validate_ui_objects()

    def __do_bind(self):
        self.Bind(
            wx.EVT_BUTTON,
            self.on_select_all,
            id=self.button_select_all.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_deselect_all,
            id=self.button_deselect_all.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_time_series_select_all,
            id=self.button_select_data["Charting Variables"]["Select"].GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_time_series_deselect_all,
            id=self.button_select_data["Charting Variables"][
                "Deselect"
            ].GetId(),
        )

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_main_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_data = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, "Data Selection"), wx.VERTICAL
        )
        sizer_time_series = wx.BoxSizer(wx.VERTICAL)
        sizer_time_series_listctrl = wx.BoxSizer(wx.HORIZONTAL)
        sizer_time_series_checkboxes = wx.BoxSizer(wx.HORIZONTAL)
        sizer_time_series_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_radbio = wx.BoxSizer(wx.VERTICAL)
        sizer_endpoints = wx.BoxSizer(wx.VERTICAL)
        sizer_dvhs = wx.BoxSizer(wx.VERTICAL)
        sizer_dvhs_checkboxes = wx.BoxSizer(wx.HORIZONTAL)

        keys = ["DVHs", "Endpoints", "Radbio"]
        static_line = {key: wx.StaticLine(self, wx.ID_ANY) for key in keys}

        sizer_dvhs_checkboxes.Add(
            self.checkbox["DVHs"], 1, wx.ALL | wx.EXPAND, 5
        )
        sizer_dvhs_checkboxes.Add(
            self.checkbox["DVHs Summary"], 1, wx.ALL | wx.EXPAND, 5
        )
        sizer_dvhs.Add(sizer_dvhs_checkboxes, 1, wx.EXPAND, 0)
        sizer_dvhs.Add(static_line["DVHs"], 0, wx.EXPAND | wx.TOP, 5)
        sizer_data.Add(sizer_dvhs, 0, wx.ALL | wx.EXPAND, 5)

        sizer_endpoints.Add(self.checkbox["Endpoints"], 0, wx.ALL, 5)
        sizer_endpoints.Add(static_line["Endpoints"], 0, wx.EXPAND | wx.TOP, 5)
        sizer_data.Add(sizer_endpoints, 0, wx.ALL | wx.EXPAND, 5)

        sizer_radbio.Add(self.checkbox["Radbio"], 0, wx.ALL, 5)
        sizer_radbio.Add(static_line["Radbio"], 0, wx.EXPAND | wx.TOP, 5)
        sizer_data.Add(sizer_radbio, 0, wx.ALL | wx.EXPAND, 5)

        sizer_time_series_checkboxes.Add(
            self.checkbox["Charting Variables"], 1, wx.EXPAND, 0
        )
        sizer_time_series_buttons.Add(
            self.button_select_data["Charting Variables"]["Select"],
            0,
            wx.ALL | wx.EXPAND,
            5,
        )
        sizer_time_series_buttons.Add(
            self.button_select_data["Charting Variables"]["Deselect"],
            0,
            wx.ALL | wx.EXPAND,
            5,
        )
        sizer_time_series_checkboxes.Add(
            sizer_time_series_buttons, 1, wx.EXPAND, 0
        )
        sizer_time_series.Add(
            sizer_time_series_checkboxes, 0, wx.ALL | wx.EXPAND, 5
        )
        sizer_time_series_listctrl.Add((20, 20), 0, 0, 0)
        sizer_time_series_listctrl.Add(
            self.list_ctrl["Charting Variables"], 1, wx.ALL | wx.EXPAND, 5
        )
        sizer_time_series.Add(
            sizer_time_series_listctrl, 0, wx.ALL | wx.EXPAND, 5
        )
        sizer_data.Add(sizer_time_series, 0, wx.ALL | wx.EXPAND, 5)

        sizer_main.Add(sizer_data, 0, wx.ALL | wx.EXPAND, 5)

        sizer_main_buttons.Add(self.button_select_all, 0, wx.ALL, 5)
        sizer_main_buttons.Add(self.button_deselect_all, 0, wx.ALL, 5)

        sizer_main_buttons.Add(self.button_save, 0, wx.ALL, 5)
        sizer_main_buttons.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_main.Add(sizer_main_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        sizer_wrapper.Add(sizer_main, 1, wx.EXPAND, 0)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def set_checkbox_values(self, value):
        for checkbox in self.checkbox.values():
            checkbox.SetValue(value)

    def on_select_all(self, evt):
        self.set_checkbox_values(True)
        self.validate_ui_objects(allow_enable=False)

    def on_deselect_all(self, evt):
        self.set_checkbox_values(False)
        self.validate_ui_objects(allow_enable=False)

    def on_time_series_select_all(self, evt):
        self.data_table_time_series.apply_selection_to_all(1)

    def on_time_series_deselect_all(self, evt):
        self.data_table_time_series.apply_selection_to_all(0)

    def validate_ui_objects(self, allow_enable=True):
        tables = {"Charting Variables": self.data_table_time_series}
        for key, data_table in tables.items():
            state = data_table.has_data
            if not state or (state and allow_enable):
                self.list_ctrl[key].Enable(state)
                self.button_select_data[key]["Select"].Enable(state)
                self.button_select_data[key]["Deselect"].Enable(state)

        for key, value in self.enabled.items():
            if not value or (value and allow_enable):
                self.checkbox[key].SetValue(value)
                self.checkbox[key].Enable(value)

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            save_data_to_file(self, "Export CSV Data", self.csv)
        self.Destroy()

    def is_checked(self, key):
        return self.checkbox[key].GetValue()

    def on_dvh_check(self, evt):
        if not self.is_checked("DVHs"):
            self.checkbox["DVHs Summary"].SetValue(False)

    @property
    def csv(self):
        csv_data = []

        csv_key = ["DVHs", "Endpoints", "Radbio", "Charting Variables"]
        csv_obj = [
            None,
            self.app.endpoint,
            self.app.radbio,
            self.app.time_series,
            self.app.control_chart,
        ]
        for i, key in enumerate(csv_key):
            if self.is_checked(key) or (
                key == "DVHs" and self.is_checked("DVHs Summary")
            ):
                csv_data.append("%s\n" % key)
                if (
                    key == "DVHs"
                ):  # DVHs has a summary and plot data for export
                    csv_data.append(
                        self.app.plot.get_csv(
                            include_summary=self.is_checked("DVHs Summary"),
                            include_dvhs=self.is_checked("DVHs"),
                        )
                    )
                else:
                    if key == "Charting Variables":
                        selection_indices = get_selected_listctrl_items(
                            self.list_ctrl["Charting Variables"]
                        )
                        y_choices = (
                            self.app.time_series.combo_box_y_axis.GetItems()
                        )
                        selection = [
                            y
                            for i, y in enumerate(y_choices)
                            if i in selection_indices
                        ]
                    else:
                        selection = None
                    csv_data.append(csv_obj[i].get_csv(selection=selection))
                csv_data.append("\n\n")

        return "\n".join(csv_data)


class ExportFigure(wx.Frame):
    def __init__(self, parent):
        wx.Frame.__init__(
            self, parent, style=wx.SYSTEM_MENU | wx.CLOSE_BOX | wx.CAPTION
        )

        self.parent = parent
        self.options = parent.options
        self.plots = {
            "DVHs": parent.plot,
            "Time Series": parent.time_series.plot,
            "Regression": parent.regression.plot,
            "Control Chart": parent.control_chart.plot,
        }

        button_keys = {"Export": wx.ID_ANY, "Dismiss": wx.ID_CANCEL}
        self.button = {
            key: wx.Button(self, button_id, key)
            for key, button_id in button_keys.items()
        }

        self.__set_input_widgets()
        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.getter = {
            int: self.get_text_ctrl_int,
            float: self.get_text_ctrl_float,
            str: self.get_combo_box,
        }

        set_msw_background_color(self)
        set_frame_icon(self)

    def __set_input_widgets(self):
        self.input = {key: [] for key in self.options.save_fig_param.keys()}
        self.label = {key: [] for key in self.options.save_fig_param.keys()}
        self.text_ctrl = {}
        self.combo_box = {}
        for obj_type, attr_dict in self.options.save_fig_param.items():
            for attr, value in attr_dict.items():
                self.label[obj_type].append(
                    wx.StaticText(
                        self, wx.ID_ANY, attr.replace("_", " ").title() + ":"
                    )
                )
                if type(value) is str:
                    color_options = ["none"] + list(plot_colors.cnames)
                    self.input[obj_type].append(
                        wx.ComboBox(
                            self,
                            wx.ID_ANY,
                            choices=color_options,
                            style=wx.CB_DROPDOWN | wx.TE_READONLY,
                        )
                    )
                    self.input[obj_type][-1].SetValue(value)
                    self.combo_box[obj_type + "_" + attr] = self.input[
                        obj_type
                    ][-1]
                else:
                    if "alpha" in attr:
                        self.input[obj_type].append(
                            wx.SpinCtrlDouble(
                                self,
                                wx.ID_ANY,
                                "0",
                                min=0,
                                max=1,
                                style=wx.SP_ARROW_KEYS,
                            )
                        )
                        self.input[obj_type][-1].SetIncrement(0.1)
                        self.input[obj_type][-1].SetValue(str(value))
                    elif "width" in attr and obj_type == "legend":
                        self.input[obj_type].append(
                            wx.SpinCtrl(
                                self,
                                wx.ID_ANY,
                                "0",
                                min=0,
                                max=20,
                                style=wx.SP_ARROW_KEYS,
                            )
                        )
                        self.input[obj_type][-1].SetValue(str(value))
                    else:
                        self.input[obj_type].append(
                            wx.TextCtrl(self, wx.ID_ANY, str(value))
                        )
                    self.text_ctrl[obj_type + "_" + attr] = self.input[
                        obj_type
                    ][-1]

        self.label_plot = wx.StaticText(self, wx.ID_ANY, "Plot:")
        self.combo_plot = wx.ComboBox(
            self, wx.ID_ANY, style=wx.CB_DROPDOWN | wx.TE_READONLY
        )

        self.include_range = wx.CheckBox(self, wx.ID_ANY, "Apply Range Edits")

    def __set_properties(self):
        self.SetTitle("Export Figure")

        self.combo_plot.SetItems(sorted(list(self.plots)))
        self.combo_plot.SetValue("DVHs")

        range_init = (
            self.options.apply_range_edits
            if hasattr(self.options, "apply_range_edits")
            else False
        )
        self.include_range.SetValue(range_init)
        self.include_range.SetToolTip(
            "Check this to alter the ranges from the current view. Leave a range field blank "
            "to use the current view's value.\n"
            "NOTE: These range edits do not apply to Machine Learning plot saves."
        )
        self.on_checkbox()  # Disable Range input objects by default

    def __do_bind(self):
        self.Bind(
            wx.EVT_BUTTON, self.on_export, id=self.button["Export"].GetId()
        )
        self.Bind(wx.EVT_BUTTON, self.on_dismiss, id=wx.ID_CANCEL)
        self.Bind(wx.EVT_CLOSE, self.on_close)
        self.Bind(
            wx.EVT_CHECKBOX, self.on_checkbox, id=self.include_range.GetId()
        )

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_input = {
            key: wx.StaticBoxSizer(
                wx.StaticBox(self, wx.ID_ANY, key.capitalize()), wx.VERTICAL
            )
            for key in ["figure", "legend"]
        }
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)

        sizer_input["figure"].Add(self.include_range, 0, 0, 0)

        for obj_type in ["figure", "legend"]:
            for i, input_obj in enumerate(self.input[obj_type]):
                sizer_input[obj_type].Add(
                    self.label[obj_type][i],
                    0,
                    wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP,
                    3,
                )
                sizer_input[obj_type].Add(
                    input_obj, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 3
                )
            sizer_main.Add(sizer_input[obj_type], 0, wx.EXPAND | wx.ALL, 5)

        sizer_main.Add(self.label_plot, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 10)
        sizer_main.Add(self.combo_plot, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 10)

        sizer_buttons.Add(
            self.button["Export"], 1, wx.EXPAND | wx.ALIGN_CENTER | wx.ALL, 5
        )
        sizer_buttons.Add(
            self.button["Dismiss"], 1, wx.EXPAND | wx.ALIGN_CENTER | wx.ALL, 5
        )
        sizer_main.Add(
            sizer_buttons, 1, wx.EXPAND | wx.ALIGN_CENTER | wx.ALL, 5
        )

        sizer_wrapper.Add(sizer_main, 0, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.SetMinSize(self.GetSize())
        self.SetMaxSize(self.GetSize())
        self.Layout()

        self.options.apply_window_position(self, "export_figure")

    @property
    def plot(self):
        return self.plots[self.combo_plot.GetValue()]

    @property
    def save_plot_function(self):
        return partial(self.plot.save_figure, self.attr_dicts)

    def on_export(self, *evt):
        self.validate_input()
        self.plot.save_figure_dlg(
            self,
            "Save %s Figure" % self.combo_plot.GetValue(),
            attr_dicts=self.attr_dicts,
        )

    def get_text_ctrl_float(self, key):
        value = self.text_ctrl[key].GetValue()
        try:
            return float(value)
        except ValueError:
            return None

    def get_text_ctrl_int(self, key):
        value = self.text_ctrl[key].GetValue()
        try:
            return int(float(value))
        except ValueError:
            return None

    def get_combo_box(self, key):
        return self.combo_box[key].GetValue()

    def get_attr_dict(self, obj_type, save_mode=False):
        return {
            key: self.getter[type(value)](obj_type + "_" + key)
            for key, value in self.options.save_fig_param[obj_type].items()
            if save_mode
            or (
                "range" not in key
                or ("range" in key and self.include_range.GetValue())
            )
        }

    @property
    def attr_dicts(self):
        return {key: self.get_attr_dict(key) for key in self.input.keys()}

    @property
    def save_attr_dicts(self):
        return {
            key: self.get_attr_dict(key, save_mode=True)
            for key in self.input.keys()
            if self.get_attr_dict(key, save_mode=True) is not None
        }

    def on_dismiss(self, *evt):
        wx.CallAfter(self.on_close)

    def on_close(self, *evt):
        self.validate_input()
        self.options.save_window_position(self, "export_figure")
        self.save_options()
        self.parent.export_figure = None
        self.Destroy()

    def save_options(self):
        self.options.save_fig_param = self.save_attr_dicts
        self.options.apply_range_edits = self.include_range.GetValue()
        self.options.save()

    def on_checkbox(self, *evt):
        for i, obj in enumerate(self.input["figure"]):
            if "Range" in self.label["figure"][i].GetLabel():
                self.label["figure"][i].Enable(self.include_range.GetValue())
                obj.Enable(self.include_range.GetValue())

    def validate_input(self):
        """If any TextCtrl is invalid, set to stored options"""
        for key, obj in self.text_ctrl.items():
            if obj.GetValue() == "":
                obj_type, attr = key[: key.find("_")], key[key.find("_") + 1 :]
                stored_value = self.options.save_fig_param[obj_type][attr]
                new_value = self.getter[type(stored_value)](key)
                if new_value is None:
                    obj.SetValue(str(stored_value))


class ExportPGSQLProgressFrame(ProgressFrame):
    """Create a window to display value generation progress and begin SaveWorker"""

    def __init__(self, file_path, export_to_json=False):
        func_call = self.func_call_json if export_to_json else self.func_call
        ProgressFrame.__init__(
            self, [file_path], func_call, title="Exporting PGSQL DB to SQLite"
        )

    def func_call(self, file_path):
        with DVH_SQL() as cnx:
            cnx.export_to_sqlite(file_path, callback=self.callback)

    def func_call_json(self, file_path):
        with DVH_SQL() as cnx:
            cnx.save_to_json(file_path, callback=self.callback)

    @staticmethod
    def callback(table, iteration, total_count):
        msg = {
            "label": "Exporting Table: %s (%s of %s)"
            % (table, iteration + 1, total_count),
            "gauge": (float(iteration) / total_count),
        }
        wx.CallAfter(pub.sendMessage, "progress_update", msg=msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.roi_geometry.py
"""
Tools for geometric calculations
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from scipy.spatial.distance import cdist
import numpy as np
from math import ceil
from shapely.geometry import Point
from dvha.tools.roi_formatter import (
    points_to_shapely_polygon,
    dicompyler_roi_to_sets_of_points,
    get_shapely_from_sets_of_points,
)


# "sets of points" objects
#     dictionaries using str(z) as keys
#         where z is the slice or z in DICOM coordinates
#         each item is a list of points representing a polygon,
#         each point is a 3-item list [x, y, z]


def union(rois):
    """Calculate the geometric union of the provided rois

    Parameters
    ----------
    rois : list
        rois formatted as "sets of points" dictionaries

    Returns
    -------
    dict
        a "sets of points" dictionary representing the union of the rois

    """

    new_roi = {}

    all_z_values = []
    for roi in rois:
        for z in list(roi):
            if z not in all_z_values:
                all_z_values.append(z)

    for z in all_z_values:

        if z not in list(new_roi):
            new_roi[z] = []

        # Convert to shapely objects
        current_slice = None
        for roi in rois:
            # Make sure current roi has at least 3 points in z plane
            if z in list(roi) and len(roi[z][0]) > 2:
                if not current_slice:
                    current_slice = points_to_shapely_polygon(roi[z])
                else:
                    current_slice = current_slice.union(
                        points_to_shapely_polygon(roi[z])
                    )

        if current_slice:
            if current_slice.type != "MultiPolygon":
                current_slice = [current_slice]

            for polygon in current_slice:
                xy = polygon.exterior.xy
                x_coord = xy[0]
                y_coord = xy[1]
                points = []
                for i in range(len(x_coord)):
                    points.append([x_coord[i], y_coord[i], round(float(z), 2)])
                new_roi[z].append(points)

                if hasattr(polygon, "interiors"):
                    for interior in polygon.interiors:
                        xy = interior.coords.xy
                        x_coord = xy[0]
                        y_coord = xy[1]
                        points = []
                        for i in range(len(x_coord)):
                            points.append(
                                [x_coord[i], y_coord[i], round(float(z), 2)]
                            )
                        new_roi[z].append(points)
        else:
            # print('WARNING: no contour found for slice %s' % z)
            pass
    return new_roi


def min_distances_to_target(oar_coordinates, target_coordinates, factors=None):
    """Calculate all OAR-point-to-Target-point euclidean distances

    Parameters
    ----------
    oar_coordinates : list
        numpy arrays of 3D points defining the surface of the OAR
    target_coordinates : list
        numpy arrays of 3D points defining the surface of the PTV
    factors :
         (Default value = None)

    Returns
    -------
    list
        min_distances: all minimum distances (cm) of OAR-point-to-Target-point pairs

    """
    # TODO: This very computationally expensive, needs a sampling method prior to calling cdist
    min_distances = []
    all_distances = cdist(oar_coordinates, target_coordinates, "euclidean")
    for i, oar_point in enumerate(all_distances):
        min_distances.append(float(np.min(oar_point) / 10.0))
        if factors is not None:
            min_distances[i] *= factors[i]

    return min_distances


def is_point_inside_roi(point, roi):
    """Check if a point is within an ROI

    Parameters
    ----------
    point : list
        x, y, z
    roi : dict
        roi: a "sets of points" formatted dictionary

    Returns
    -------
    bool
        Whether or not the poin is within the roi

    """
    z_keys = list(roi.keys())
    roi_z = np.array([float(z) for z in z_keys])
    if np.max(roi_z) > point[2] > np.min(roi_z):
        nearest_z_index = (np.abs(roi_z - point[2])).argmin()
        nearest_z_key = z_keys[nearest_z_index]
        if (
            abs(float(nearest_z_key) - point[2]) < 0.5
        ):  # make sure point is within 0.5mm
            if (
                len(roi[nearest_z_key]) > 2
            ):  # make sure there are 3 points to make a polygon
                shapely_roi = points_to_shapely_polygon(roi[nearest_z_key])
                shapely_point = Point(point[0], point[1])
                return shapely_point.within(shapely_roi)
    return False


def cross_section(roi):
    """Calculate the cross section of a given roi

    Parameters
    ----------
    roi : dict
        a "sets of points" formatted dictionary

    Returns
    -------
    dict
        max and median cross-sectional area of all slices in cm^2

    """
    areas = []

    for z in list(roi):
        shapely_roi = points_to_shapely_polygon(roi[z])
        if shapely_roi and shapely_roi.area > 0:
            slice_centroid = shapely_roi.centroid
            polygon_count = len(slice_centroid.xy[0])
            for i in range(polygon_count):
                if polygon_count > 1:
                    areas.append(shapely_roi[i].area)
                else:
                    areas.append(shapely_roi.area)

    areas = np.array(areas)

    area = {
        "max": float(np.max(areas) / 100.0),
        "median": float(np.median(areas) / 100.0),
    }

    return area


def surface_area(coord, coord_type="dicompyler"):
    """Calculate the surface of a given roi

    Parameters
    ----------
    coord :
        dicompyler structure coordinates from GetStructureCoordinates() or a sets_of_points dictionary
    coord_type :
        either 'dicompyler' or 'sets_of_points' (Default value = 'dicompyler')

    Returns
    -------
    float
        surface_area in cm^2

    """
    # TODO: This surface area method needs validation, but likely needs to be corrected

    if coord_type == "sets_of_points":
        sets_of_points = coord
    else:
        sets_of_points = dicompyler_roi_to_sets_of_points(coord)

    shapely_roi = get_shapely_from_sets_of_points(sets_of_points)

    slice_count = len(shapely_roi["z"])

    area = 0.0
    polygon = shapely_roi["polygon"]
    z = shapely_roi["z"]
    thickness = min(shapely_roi["thickness"])

    for i in range(slice_count):
        for j in [-1, 1]:  # -1 for bottom area and 1 for top area
            # ensure bottom of first slice and top of last slice are fully added
            # if prev/next slice is not adjacent, assume non-contiguous ROI
            if (
                (i == 0 and j == -1)
                or (i == slice_count - 1 and j == 1)
                or abs(z[i] - z[i + j]) > 2 * thickness
            ):
                area += polygon[i].area
            else:
                area += polygon[i].difference(polygon[i + j]).area

        area += polygon[i].length * thickness

    return round(area / 100, 3)


def overlap_volume(oar, tv):
    """Calculate the overlap volume of two rois

    Parameters
    ----------
    oar : dict
        organ-at-risk as a "sets of points" formatted dictionary
    tv : dict
        treatment volume as a "sets of points" formatted dictionary

    Returns
    -------

    """

    intersection_volume = 0.0
    all_z_values = [round(float(z), 2) for z in list(tv)]
    all_z_values = np.sort(all_z_values)
    thicknesses = np.abs(np.diff(all_z_values))
    if len(thicknesses) > 0:
        thicknesses = np.append(thicknesses, np.min(thicknesses))
        all_z_values = all_z_values.tolist()

        for z in list(tv):
            # z in coord will not necessarily go in order of z, convert z to float to lookup thickness
            # also used to check for top and bottom slices, to add area of those contours

            if z in list(oar):
                thickness = thicknesses[all_z_values.index(round(float(z), 2))]
                shapely_tv = points_to_shapely_polygon(tv[z])
                shapely_oar = points_to_shapely_polygon(oar[z])
                if shapely_oar and shapely_tv:
                    intersection_volume += (
                        shapely_tv.intersection(shapely_oar).area * thickness
                    )

        return round(intersection_volume / 1000.0, 2)
    return 0.0


def volume(roi):
    """

    Parameters
    ----------
    roi :
        a "sets of points" formatted dictionary

    Returns
    -------
    float
        volume in cm^3 of roi

    """

    # oar and ptv are lists using str(z) as keys
    # each item is an ordered list of points representing a polygon
    # polygon n is inside polygon n-1, then the current accumulated polygon is
    #    polygon n subtracted from the accumulated polygon up to and including polygon n-1
    #    Same method DICOM uses to handle rings and islands

    vol = 0.0
    all_z_values = [round(float(z), 2) for z in list(roi)]
    all_z_values = np.sort(all_z_values)
    thicknesses = np.abs(np.diff(all_z_values))
    thicknesses = np.append(thicknesses, np.min(thicknesses))
    all_z_values = all_z_values.tolist()

    for z in list(roi):
        # z in coord will not necessarily go in order of z, convert z to float to lookup thickness
        # also used to check for top and bottom slices, to add area of those contours

        thickness = thicknesses[all_z_values.index(round(float(z), 2))]
        shapely_roi = points_to_shapely_polygon(roi[z])
        if shapely_roi:
            vol += shapely_roi.area * thickness

    return round(vol / 1000.0, 2)


def centroid(roi):
    """

    Parameters
    ----------
    roi :
        a "sets of points" formatted dictionary

    Returns
    -------
    list
        centroid or the roi in x, y, z dicom coordinates (mm)

    """
    centroids = {"x": [], "y": [], "z": [], "area": []}

    for z in list(roi):
        shapely_roi = points_to_shapely_polygon(roi[z])
        if shapely_roi and shapely_roi.area > 0:
            slice_centroid = shapely_roi.centroid
            polygon_count = len(slice_centroid.xy[0])
            for i in range(polygon_count):
                centroids["x"].append(slice_centroid.xy[0][i])
                centroids["y"].append(slice_centroid.xy[1][i])
                centroids["z"].append(float(z))
                if polygon_count > 1:
                    centroids["area"].append(shapely_roi[i].area)
                else:
                    centroids["area"].append(shapely_roi.area)

    x = np.array(centroids["x"])
    y = np.array(centroids["y"])
    z = np.array(centroids["z"])
    w = np.array(centroids["area"])
    w_sum = np.sum(w)

    volumetric_centroid = [
        float(np.sum(x * w) / w_sum),
        float(np.sum(y * w) / w_sum),
        float(np.sum(z * w) / w_sum),
    ]

    return volumetric_centroid


def spread(roi):
    """

    Parameters
    ----------
    roi :
        a "sets of points" formatted dictionary

    Returns
    -------
    list
        x, y, z dimensions of a rectangular prism encompassing roi

    """
    all_points = {"x": [], "y": [], "z": []}

    for z in list(roi):
        for polygon in roi[z]:
            for point in polygon:
                all_points["x"].append(point[0])
                all_points["y"].append(point[1])
                all_points["z"].append(point[2])

    all_points = {
        "x": np.array(all_points["x"]),
        "y": np.array(all_points["y"]),
        "z": np.array(all_points["z"]),
    }

    if len(all_points["x"] > 1):
        data = [
            abs(float(np.max(all_points[dim]) - np.min(all_points[dim])))
            / 10.0
            for dim in ["x", "y", "z"]
        ]
    else:
        data = [0, 0, 0]

    return data


def dth(min_distances):
    """

    Parameters
    ----------
    min_distances :
        the output from min_distances_to_target

    Returns
    -------
    numpy.array
        histogram of distances in 1mm bin widths

    """
    min_distances = 10.0 * np.array(min_distances)
    max_abs_value = int(ceil(np.max(np.abs(min_distances))))
    data, bins = np.histogram(
        min_distances,
        bins=max_abs_value * 2 + 1,
        range=(-max_abs_value - 1, max_abs_value),
    )

    return np.divide(data, np.sum(data))


def process_dth_string(dth_string):
    """Convert a dth_string from the database into data and bins
    DVHA stores 1-mm binned surface DTHs with an odd number of bins, middle bin is 0.

    Parameters
    ----------
    dth_string :
        a value from the dth_string column

    Returns
    -------
    type
        counts, bin positions (mm)

    """
    counts = np.array(dth_string.split(","), dtype=np.float)
    max_bin = (len(counts) - 1) / 2
    bins = np.linspace(-max_bin, max_bin, len(counts))
    return bins, counts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# setup.py
"""
A setuptools setup file for DVH Analytics
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import os
from setuptools import setup, find_packages
from dvha._version import __version__


if os.environ.get("READTHEDOCS") == "True":
    requirements_path = "docs/requirements.txt"
else:
    requirements_path = "requirements.txt"

with open(requirements_path, "r") as doc:
    requires = [line.strip() for line in doc]

with open("README.rst", "r") as doc:
    long_description = doc.read()


setup(
    name="dvha",
    include_package_data=True,
    python_requires=">3.5",
    packages=find_packages(),
    version=__version__,
    description="Create a database of DVHs, GUI with wxPython, "
                "plots with Bokeh",
    author="Dan Cutright",
    author_email="dan.cutright@gmail.com",
    url="https://github.com/cutright/DVH-Analytics-Desktop",
    download_url="https://github.com/cutright/DVH-Analytics-Desktop/archive/master.zip",
    license="BSD License",
    keywords=[
        "dvh",
        "radiation therapy",
        "research",
        "dicom",
        "dicom-rt",
        "bokeh",
        "analytics",
        "wxpython",
    ],
    classifiers=[],
    install_requires=requires,
    entry_points={"console_scripts": ["dvha = dvha.main:start"]},
    long_description=long_description,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.name_prediction.py
"""Implementation of rapidfuzz for ROI name prediction"""
# Copyright (c) 2016-2021 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics
from rapidfuzz import fuzz
from dvha.tools.roi_name_manager import clean_name


class ROINamePredictor:
    """ROI Name Prediction class object

    Parameters
    ----------
    roi_map : DatabaseROIs
        ROI map object
    weight_simple : float, optional
        Scaling factor for fuzz.ratio for combined score
    weight_partial : float, optional
        Scaling factor for fuzz.partial_ratio for combined score
    threshold : float, optional
        Set a minimum score for a prediction to be returned
    """

    def __init__(
        self, roi_map, weight_simple=1.0, weight_partial=0.6, threshold=0.0
    ):

        self.roi_map = roi_map
        norm_weight = weight_partial + weight_simple
        self.weight = {
            "simple": 2 * weight_simple / norm_weight,
            "partial": 2 * weight_partial / norm_weight,
        }
        self.threshold = threshold

    def get_best_roi_match(self, roi, physician, return_score=False):
        """Check all ROI variations for best match, return physician ROI

        Parameters
        ----------
        roi : str
            An ROI name
        physician : str
            Physician as stored in ROI Map
        return_score : bool, optional
             If true, return a tuple: prediction, score

        Returns
        -------
        str
            The physician ROI associated with the ROI variation that is has
            the highest combined fuzz score for ``roi``

        """
        physician_variations = self.roi_map.get_all_variations_of_physician(
            physician
        )
        fuzz_scores = self.get_combined_fuzz_scores(roi, physician_variations)
        if fuzz_scores:
            predicted_variation, score = fuzz_scores[0][1], fuzz_scores[0][0]
            prediction = self.roi_map.get_physician_roi(
                physician, predicted_variation
            )
            if score > self.threshold:
                if return_score:
                    return prediction, score
                return prediction

    def get_combined_fuzz_score(self, a, b, mode="geom_mean"):
        """Return ``combine_scores`` for strings ``a`` and ``b``

        Parameters
        ----------
        a : str
            Any string
        b : str
            Another string for comparison
        mode : str, optional
            Method for combining ``fuzz.ratio`` and ``fuzz.partial_ratio``.
            Options are 'geom_mean', 'product', and 'average'

        Returns
        -------
        float
            Results from ``combine_scores`` for ``a`` and ``b``

        """
        a, b = clean_name(a), clean_name(b)

        simple = float(fuzz.ratio(a, b) * self.weight["simple"])
        partial = float(fuzz.partial_ratio(a, b) * self.weight["partial"])

        return self.combine_scores(simple, partial, mode=mode)

    @staticmethod
    def combine_scores(score_1, score_2, mode="average"):
        """Get a combined fuzz score

        Parameters
        ----------
        score_1 : float
            A fuzz ratio score
        score_2 : float
            Another fuzz ratio score
        mode : str, optional
            Method for combining ``score_1`` and ``score_2``.
            Options are 'geom_mean', 'product', and 'average'

        Returns
        -------
        float
            Combined score

        """
        if mode == "geom_mean":
            return (score_1 * score_2) ** 0.5
        elif mode == "product":
            return score_1 * score_2 / 100.0
        else:  # average
            return (score_1 + score_2) / 2.0

    def get_combined_fuzz_scores(self, string, list_of_strings):
        """Compare a string against many

        Parameters
        ----------
        string : str
            A string to compare against each string in ``list_of_strings``
        list_of_strings : list
            A list of strings for comparison

        Returns
        -------
        list
            A list of tuples (score, string) in order of score

        """
        scores = [
            self.get_combined_fuzz_score(string, string_b)
            for string_b in list_of_strings
        ]
        if scores:
            order_index = sorted(range(len(scores)), key=lambda k: scores[k])
            return [(scores[i], list_of_strings[i]) for i in order_index[::-1]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.machine_learning.py
"""
Classes to view and calculate Machine Learning predictions
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
import numpy as np
from sklearn.ensemble import (
    RandomForestRegressor,
    GradientBoostingRegressor,
    RandomForestClassifier,
    GradientBoostingClassifier,
)
from sklearn.model_selection import train_test_split
from sklearn.svm import SVR, SVC
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.neural_network import MLPClassifier, MLPRegressor
from dvha.dialogs.export import save_data_to_file
from dvha.dialogs.main import ShowList
from dvha.options import DefaultOptions
from dvha.paths import MODELS_DIR
from dvha.models.plot import PlotMachineLearning, PlotFeatureImportance
from dvha.tools.utilities import (
    set_msw_background_color,
    get_window_size,
    load_object_from_file,
    set_frame_icon,
    get_selected_listctrl_items,
)


class MachineLearningFrame(wx.Frame):
    def __init__(
        self,
        main_app_frame,
        data,
        title,
        sklearn_predictor=None,
        alg_type="regressor",
        tool_tips=None,
        include_test_data=True,
    ):
        wx.Frame.__init__(self, None)

        self.main_app_frame = main_app_frame
        self.data = data
        self.title = title
        self.sklearn_predictor = [
            sklearn_predictor,
            ALGORITHMS[title][alg_type],
        ][sklearn_predictor is None]
        self.tool_tips = [tool_tips, ALGORITHMS[title]["tool_tips"]][
            tool_tips is None
        ]
        self.include_test_data = include_test_data

        self.model = None
        is_classifier = alg_type == "classifier"
        self.plot = PlotMachineLearning(
            self,
            ml_type=self.title,
            ml_type_short=self.ml_type_short,
            include_test_data=include_test_data,
            is_classifier=is_classifier,
            **self.data
        )

        self.feature_importance_dlg = None

        self.input = {}
        self.defaults = {}
        self.getters = {}

        self.data_split_input = {
            "test_size": wx.TextCtrl(self, wx.ID_ANY, "0.25"),
            "train_size": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "random_state": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "shuffle": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
        }

        self.data_split_defaults = {
            "test_size": 0.25,
            "train_size": None,
            "random_state": None,
            "shuffle": True,
        }

        self.data_split_getters = {
            "test_size": self.to_float_or_none,
            "train_size": self.to_float_or_none,
            "random_state": self.to_int_or_none,
            "shuffle": self.to_bool,
        }

        self.button_calculate = wx.Button(self, wx.ID_ANY, "Calculate")
        self.button_features = wx.Button(self, wx.ID_ANY, "Features")
        self.button_importance = wx.Button(self, wx.ID_ANY, "Importance Plot")
        self.button_export_data = wx.Button(self, wx.ID_ANY, "Export Data")
        self.button_save_figure = wx.Button(self, wx.ID_ANY, "Save Figure")
        self.button_save_model = wx.Button(self, wx.ID_ANY, "Save Model")

        self.do_bind()

    def do_bind(self):
        self.Bind(
            wx.EVT_BUTTON, self.on_calculate, id=self.button_calculate.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON, self.on_features, id=self.button_features.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_feature_importance,
            id=self.button_importance.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON, self.on_export, id=self.button_export_data.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_save_figure,
            id=self.button_save_figure.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_save_model,
            id=self.button_save_model.GetId(),
        )
        self.Bind(wx.EVT_SIZE, self.on_resize)

    def set_properties(self):
        self.SetTitle(self.title)
        x_size = [0.6, 0.8][self.include_test_data]
        self.SetMinSize(get_window_size(x_size, 0.7))
        self.set_defaults()

        for key, input_obj in self.input.items():
            input_obj.SetToolTip(self.tool_tips[key])

        for key, input_obj in self.data_split_input.items():
            input_obj.SetToolTip(DATA_SPLIT_TOOL_TIPS[key])

    def do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.HORIZONTAL)
        sizer_side_bar = wx.BoxSizer(wx.VERTICAL)
        sizer_actions = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, "Actions"), wx.VERTICAL
        )
        sizer_param = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, "Parameters"), wx.VERTICAL
        )
        sizer_split_param = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, "Data Split"), wx.VERTICAL
        )

        variables = list(self.input)
        variables.sort()
        sizer_input = {
            variable: wx.BoxSizer(wx.HORIZONTAL) for variable in variables
        }
        for variable in variables:
            sizer_input[variable].Add(
                wx.StaticText(self, wx.ID_ANY, "%s:\t" % variable),
                0,
                wx.EXPAND,
                0,
            )
            sizer_input[variable].Add(self.input[variable], 1, wx.EXPAND, 0)
            sizer_param.Add(sizer_input[variable], 1, wx.EXPAND | wx.ALL, 2)
        sizer_side_bar.Add(sizer_param, 0, wx.ALL | wx.EXPAND, 5)

        split_variables = [
            "test_size",
            "train_size",
            "random_state",
            "shuffle",
        ]
        sizer_split_input = {
            variable: wx.BoxSizer(wx.HORIZONTAL)
            for variable in split_variables
        }
        for variable in split_variables:
            sizer_split_input[variable].Add(
                wx.StaticText(self, wx.ID_ANY, "%s:\t" % variable),
                0,
                wx.EXPAND,
                0,
            )
            sizer_split_input[variable].Add(
                self.data_split_input[variable], 1, wx.EXPAND, 0
            )
            sizer_split_param.Add(
                sizer_split_input[variable], 1, wx.EXPAND | wx.ALL, 2
            )
        sizer_side_bar.Add(sizer_split_param, 0, wx.ALL | wx.EXPAND, 5)

        sizer_actions.Add(
            self.button_calculate,
            1,
            wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP,
            5,
        )
        sizer_actions.Add(
            self.button_features, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5
        )
        sizer_actions.Add(
            self.button_importance,
            1,
            wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP,
            5,
        )
        sizer_actions.Add(
            self.button_export_data, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )
        sizer_actions.Add(
            self.button_save_figure, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )
        sizer_actions.Add(
            self.button_save_model,
            1,
            wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT,
            5,
        )
        sizer_side_bar.Add(sizer_actions, 0, wx.ALL | wx.EXPAND, 5)

        sizer_wrapper.Add(sizer_side_bar, 0, wx.EXPAND, 0)

        sizer_wrapper.Add(self.plot.layout, 1, wx.EXPAND, 0)

        set_msw_background_color(
            self
        )  # If windows, change the background color
        set_frame_icon(self)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def get_param(self, variable):
        ans = variable in list(self.input)
        input_dict = [self.data_split_input, self.input][ans]
        default_dict = [self.data_split_defaults, self.defaults][ans]
        getters_dict = [self.data_split_getters, self.getters][ans]

        try:
            return getters_dict[variable](input_dict[variable].GetValue())
        except Exception:
            return default_dict[variable]

    def set_defaults(self):
        for variable, input_obj in self.input.items():
            input_obj.SetValue(self.to_str_for_gui(self.defaults[variable]))
        for variable, input_obj in self.data_split_input.items():
            input_obj.SetValue(
                self.to_str_for_gui(self.data_split_defaults[variable])
            )

    @property
    def input_parameters(self):
        return {
            variable: self.get_param(variable)
            for variable in self.input.keys()
            if self.get_param(variable) != self.defaults[variable]
        }

    def set_input_parameters(self, input_parameters):
        for variable, value in input_parameters.items():
            self.input[variable].SetValue(str(value))

    @property
    def data_split_parameters(self):
        return {
            variable: self.get_param(variable)
            for variable in self.data_split_input.keys()
            if self.get_param(variable) != self.data_split_defaults[variable]
        }

    def set_data_split_parameters(self, data_split_input):
        for variable, value in data_split_input.keys():
            self.data_split_input[variable].SetValue(str(value))

    def to_int_float_or_none(self, str_value):
        if str_value.lower() == "none":
            return None
        return self.to_int_or_float(str_value)

    @staticmethod
    def to_int_float_string_or_none(str_value):
        if str_value.lower() == "none":
            return None
        if str_value.isnumeric():  # int
            return int(float(str_value))
        if "." in str_value and len(str_value.split(".") == 2):
            return float(str_value)
        return str_value

    @staticmethod
    def to_int_or_float(str_value):
        if str_value.isnumeric():  # int
            return int(float(str_value))
        return float(str_value)

    def to_int_or_none(self, str_value):
        if str_value.lower() == "none":
            return None
        return self.to_int(str_value)

    def to_float_or_none(self, str_value):
        if str_value.lower() == "none":
            return None
        return self.to_float(str_value)

    @staticmethod
    def to_int(str_value):
        return int(float(str_value))

    @staticmethod
    def to_float(str_value):
        return float(str_value)

    @staticmethod
    def to_str(str_value):
        return str_value

    @staticmethod
    def to_float_or_str(str_value):
        if "." in str_value and len(str_value.split(".") == 2):
            if (
                "%s" % (str_value.split(".")[0] + str_value.split(".")[1])
            ).isnumeric():
                return float(str_value)
        return str_value

    def to_float_str_or_none(self, str_value):
        try:
            return self.to_float_or_none(str_value)
        except Exception:
            return str_value

    @staticmethod
    def to_bool(str_value):
        if str_value.lower() == "true":
            return True
        return False

    @staticmethod
    def to_bool_or_str(str_value):
        if str_value.lower() == "true":
            return True
        if str_value.lower() == "false":
            return False
        return str_value

    @staticmethod
    def to_str_for_gui(value):
        if value is None:
            return "None"
        return str(value)

    @property
    def plot_data(self):
        try:
            self.model = self.sklearn_predictor(**self.input_parameters)
            return MachineLearningPlotData(
                self.data["X"],
                self.data["y"],
                self.model,
                **self.data_split_parameters
            )
        except Exception as e:
            wx.MessageBox(
                str(e), "Error!", wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING
            )

    def do_prediction(self):
        self.model = self.sklearn_predictor(**self.input_parameters)
        return MachineLearningPlotData(
            self.data["X"],
            self.data["y"],
            self.model,
            **self.data_split_parameters
        )

    def on_calculate(self, evt):
        data = self.plot_data
        if data is not None:
            self.plot.update_data(data)

    def redraw_plot(self):
        self.plot.redraw_plot()

    def on_resize(self, *evt):
        try:
            self.Refresh()
            self.Layout()
            wx.CallAfter(self.redraw_plot)
        except RuntimeError:
            pass

    def on_export(self, evt):
        save_data_to_file(
            self, "Save machine learning data to csv", self.plot.get_csv()
        )

    def on_save_figure(self, *evt):
        title = "Save %s Plot to .html" % self.title.title()
        export_frame = self.main_app_frame.export_figure
        attr_dicts = None if export_frame is None else export_frame.attr_dicts
        self.plot.save_figure_dlg(self, title, attr_dicts=attr_dicts)

    def on_save_model(self, evt):
        data = {
            "y_variable": self.plot.y_variable,
            "model": self.model,
            "sklearn_predictor": self.sklearn_predictor,
            "tool_tips": self.tool_tips,
            "x_variables": self.plot.x_variables,
            "title": self.title,
            "input_parameters": self.input_parameters,
            "data_split": self.data_split_parameters,
            "version": DefaultOptions().VERSION,
        }
        save_data_to_file(
            self,
            "Save Model",
            data,
            wildcard="MODEL files (*.ml)|*.ml",
            data_type="pickle",
            initial_dir=MODELS_DIR,
        )

    def run(self):
        self.set_properties()
        self.do_layout()
        self.Show()
        self.on_calculate(None)

    @property
    def frame_size(self):
        return self.GetSize()

    @property
    def ml_type_short(self):
        return "".join([s[0] for s in self.title.split(" ")]).upper()

    def on_features(self, *evt):
        ShowList(self.plot.x_variables, "Features")

    def on_feature_importance(self, evt):
        title = "Importance Figure for %s (%s)" % (
            self.title,
            self.data["y_variable"],
        )
        plot_title = "%s Feature Importances for %s" % (
            self.title,
            self.data["y_variable"],
        )
        self.feature_importance_dlg = FeatureImportanceFrame(
            self.data["options"],
            self.data["x_variables"],
            self.model.feature_importances_,
            title,
            plot_title,
        )
        self.feature_importance_dlg.Show()


class RandomForestFrame(MachineLearningFrame):
    def __init__(
        self,
        main_app_frame,
        data,
        include_test_data=True,
        alg_type="regressor",
    ):
        tool_tips = [RF_TOOL_TIPS_CLASSIFIER, RF_TOOL_TIPS][
            alg_type == "regressor"
        ]
        crit_choices = [["gini", "entropy"], ["mse", "mae"]][
            alg_type == "regressor"
        ]
        MachineLearningFrame.__init__(
            self,
            main_app_frame,
            data,
            "Random Forest",
            include_test_data=include_test_data,
            alg_type=alg_type,
            tool_tips=tool_tips,
        )

        self.input = {
            "n_estimators": wx.TextCtrl(self, wx.ID_ANY, "100"),
            "criterion": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=crit_choices,
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "max_depth": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "min_samples_split": wx.TextCtrl(self, wx.ID_ANY, "2"),
            "min_samples_leaf": wx.TextCtrl(self, wx.ID_ANY, "1"),
            "min_weight_fraction_leaf": wx.TextCtrl(self, wx.ID_ANY, "0"),
            "max_features": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "max_leaf_nodes": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "min_impurity_decrease": wx.TextCtrl(self, wx.ID_ANY, "0"),
            "bootstrap": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "oob_score": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "n_jobs": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "random_state": wx.TextCtrl(self, wx.ID_ANY, "None"),
        }

        self.defaults = {
            "n_estimators": 100,
            "criterion": crit_choices[0],
            "max_depth": None,
            "min_samples_split": 2,
            "min_samples_leaf": 1,
            "min_weight_fraction_leaf": 0.0,
            "max_features": None,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.0,
            "bootstrap": True,
            "oob_score": False,
            "n_jobs": None,
            "random_state": None,
        }

        self.getters = {
            "n_estimators": self.to_int,
            "criterion": self.to_str,
            "max_depth": self.to_int_or_none,
            "min_samples_split": self.to_int_or_float,
            "min_samples_leaf": self.to_int_or_float,
            "min_weight_fraction_leaf": self.to_float,
            "max_features": self.to_int_float_string_or_none,
            "max_leaf_nodes": self.to_int_or_none,
            "min_impurity_decrease": self.to_float,
            "bootstrap": self.to_bool,
            "oob_score": self.to_bool,
            "n_jobs": self.to_int_or_none,
            "random_state": self.to_int_or_none,
        }

        self.run()


class GradientBoostingFrame(MachineLearningFrame):
    def __init__(
        self,
        main_app_frame,
        data,
        include_test_data=True,
        alg_type="regressor",
    ):
        MachineLearningFrame.__init__(
            self,
            main_app_frame,
            data,
            "Gradient Boosting",
            include_test_data=include_test_data,
            alg_type=alg_type,
        )

        self.input = {
            "loss": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["ls", "lad", "huber", "quantile"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "learning_rate": wx.TextCtrl(self, wx.ID_ANY, "0.1"),
            "n_estimators": wx.TextCtrl(self, wx.ID_ANY, "100"),
            "subsample": wx.TextCtrl(self, wx.ID_ANY, "1.0"),
            "criterion": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["friedman_mse", "mse", "mae"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "max_depth": wx.TextCtrl(self, wx.ID_ANY, "3"),
            "min_samples_split": wx.TextCtrl(self, wx.ID_ANY, "2"),
            "min_samples_leaf": wx.TextCtrl(self, wx.ID_ANY, "1"),
            "min_weight_fraction_leaf": wx.TextCtrl(self, wx.ID_ANY, "0"),
            "max_features": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "alpha": wx.TextCtrl(self, wx.ID_ANY, "0.9"),
            "max_leaf_nodes": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "min_impurity_decrease": wx.TextCtrl(self, wx.ID_ANY, "0"),
            "init": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["DummyEstimator", "zero"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "random_state": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "presort": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["auto", "True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "validation_fraction": wx.TextCtrl(self, wx.ID_ANY, "0.1"),
            "n_iter_no_change": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "tol": wx.TextCtrl(self, wx.ID_ANY, "1e-4"),
        }

        self.defaults = {
            "loss": "ls",
            "learning_rate": 0.1,
            "n_estimators": 100,
            "subsample": 1.0,
            "criterion": "friedman_mse",
            "max_depth": 3,
            "min_samples_split": 2,
            "min_samples_leaf": 1,
            "min_weight_fraction_leaf": 0,
            "max_features": None,
            "alpha": 0.9,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0,
            "init": "DummyEstimator",
            "random_state": None,
            "presort": "auto",
            "validation_fraction": 0.1,
            "n_iter_no_change": None,
            "tol": float("1e-4"),
        }

        self.getters = {
            "loss": self.to_str,
            "learning_rate": self.to_float,
            "n_estimators": self.to_int,
            "subsample": self.to_float,
            "criterion": self.to_str,
            "max_depth": self.to_int,
            "min_samples_split": self.to_int_or_float,
            "min_samples_leaf": self.to_int_or_float,
            "min_weight_fraction_leaf": self.to_float,
            "max_features": self.to_int_float_string_or_none,
            "alpha": self.to_float,
            "max_leaf_nodes": self.to_int_or_none,
            "min_impurity_decrease": self.to_float,
            "init": self.to_str,
            "random_state": self.to_int_or_none,
            "presort": self.to_bool_or_str,
            "validation_fraction": self.to_float,
            "n_iter_no_change": self.to_int_or_none,
            "tol": self.to_float,
        }

        self.run()


class DecisionTreeFrame(MachineLearningFrame):
    def __init__(
        self,
        main_app_frame,
        data,
        include_test_data=True,
        alg_type="regressor",
    ):
        MachineLearningFrame.__init__(
            self,
            main_app_frame,
            data,
            "Decision Tree",
            include_test_data=include_test_data,
            alg_type=alg_type,
        )

        self.input = {
            "criterion": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["mse", "friedman_mse", "mae"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "splitter": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["best", "random"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "max_depth": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "min_samples_split": wx.TextCtrl(self, wx.ID_ANY, "2"),
            "min_samples_leaf": wx.TextCtrl(self, wx.ID_ANY, "1"),
            "min_weight_fraction_leaf": wx.TextCtrl(self, wx.ID_ANY, "0"),
            "max_features": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "max_leaf_nodes": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "min_impurity_decrease": wx.TextCtrl(self, wx.ID_ANY, "0"),
            "random_state": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "presort": wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
        }

        self.defaults = {
            "criterion": "mse",
            "splitter": "best",
            "max_depth": None,
            "min_samples_split": 2,
            "min_samples_leaf": 1,
            "min_weight_fraction_leaf": 0.0,
            "max_features": None,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.0,
            "random_state": None,
            "presort": False,
        }

        self.getters = {
            "criterion": self.to_str,
            "splitter": self.to_str,
            "max_depth": self.to_int_or_none,
            "min_samples_split": self.to_int_or_float,
            "min_samples_leaf": self.to_int_or_float,
            "min_weight_fraction_leaf": self.to_float,
            "max_features": self.to_int_float_string_or_none,
            "max_leaf_nodes": self.to_int_or_none,
            "min_impurity_decrease": self.to_float,
            "random_state": self.to_int_or_none,
            "presort": self.to_bool,
        }

        self.run()


class SupportVectorRegressionFrame(MachineLearningFrame):
    def __init__(
        self,
        main_app_frame,
        data,
        include_test_data=True,
        alg_type="regressor",
    ):
        MachineLearningFrame.__init__(
            self,
            main_app_frame,
            data,
            "Support Vector Machine",
            include_test_data=include_test_data,
            alg_type=alg_type,
        )

        self.input = {
            "kernel": wx.ComboBox(
                self,
                wx.ID_ANY,
                "rbf",
                choices=["linear", "poly", "rbf", "sigmoid", "precomputed"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "degree": wx.TextCtrl(self, wx.ID_ANY, "3"),
            "gamma": wx.TextCtrl(self, wx.ID_ANY, "auto"),
            "coef0": wx.TextCtrl(self, wx.ID_ANY, "0.0"),
            "tol": wx.TextCtrl(self, wx.ID_ANY, "1e-3"),
            "C": wx.TextCtrl(self, wx.ID_ANY, "1.0"),
            "epsilon": wx.TextCtrl(self, wx.ID_ANY, "0.1"),
            "shrinking": wx.ComboBox(
                self,
                wx.ID_ANY,
                "True",
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "cache_size": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "max_iter": wx.TextCtrl(self, wx.ID_ANY, "-1"),
        }

        self.defaults = {
            "kernel": "rbf",
            "degree": 3,
            "gamma": "scale",
            "coef0": 0.0,
            "tol": 0.001,
            "C": 1.0,
            "epsilon": 0.1,
            "shrinking": True,
            "cache_size": None,
            "max_iter": -1,
        }

        self.getters = {
            "kernel": self.to_str,
            "degree": self.to_int,
            "gamma": self.to_float_or_str,
            "coef0": self.to_float,
            "tol": self.to_float,
            "C": self.to_float,
            "epsilon": self.to_float,
            "shrinking": self.to_bool,
            "cache_size": self.to_float_or_none,
            "max_iter": self.to_int,
        }

        self.button_importance.Disable()

        self.run()


class MLPFrame(MachineLearningFrame):
    def __init__(
        self,
        main_app_frame,
        data,
        include_test_data=True,
        alg_type="regressor",
    ):
        MachineLearningFrame.__init__(
            self,
            main_app_frame,
            data,
            "Multilayer Perceptron",
            include_test_data=include_test_data,
            alg_type=alg_type,
        )

        self.input = {
            "activation": wx.ComboBox(
                self,
                wx.ID_ANY,
                "relu",
                choices=["identity", "logistic", "tanh", "relu"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "solver": wx.ComboBox(
                self,
                wx.ID_ANY,
                "adam",
                choices=["lbfgs", "sgd", "adam"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "alpha": wx.TextCtrl(self, wx.ID_ANY, "0.001"),
            "batch_size": wx.TextCtrl(self, wx.ID_ANY, "auto"),
            "learning_rate": wx.ComboBox(
                self,
                wx.ID_ANY,
                "constant",
                choices=["constant", "invscaling", "adaptive"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "learning_rate_init": wx.TextCtrl(self, wx.ID_ANY, "0.001"),
            "power_t": wx.TextCtrl(self, wx.ID_ANY, "0.5"),
            "max_iter": wx.TextCtrl(self, wx.ID_ANY, "200"),
            "shuffle": wx.ComboBox(
                self,
                wx.ID_ANY,
                "True",
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "random_state": wx.TextCtrl(self, wx.ID_ANY, "None"),
            "tol": wx.TextCtrl(self, wx.ID_ANY, "1e-4"),
            "warm_start": wx.ComboBox(
                self,
                wx.ID_ANY,
                "False",
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "momentum": wx.TextCtrl(self, wx.ID_ANY, "0.9"),
            "nesterovs_momentum": wx.ComboBox(
                self,
                wx.ID_ANY,
                "True",
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "early_stopping": wx.ComboBox(
                self,
                wx.ID_ANY,
                "False",
                choices=["True", "False"],
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            ),
            "validation_fraction": wx.TextCtrl(self, wx.ID_ANY, "0.1"),
            "beta_1": wx.TextCtrl(self, wx.ID_ANY, "0.9"),
            "beta_2": wx.TextCtrl(self, wx.ID_ANY, "0.999"),
            "epsilon": wx.TextCtrl(self, wx.ID_ANY, "1e-8"),
            "n_iter_no_change": wx.TextCtrl(self, wx.ID_ANY, "10"),
        }

        self.defaults = {
            "activation": "relu",
            "solver": "adam",
            "alpha": 0.0001,
            "batch_size": "auto",
            "learning_rate": "constant",
            "learning_rate_init": 0.001,
            "power_t": 0.5,
            "max_iter": 200,
            "shuffle": True,
            "random_state": None,
            "tol": 1e-4,
            "warm_start": False,
            "momentum": 0.9,
            "nesterovs_momentum": True,
            "early_stopping": False,
            "validation_fraction": 0.1,
            "beta_1": 0.9,
            "beta_2": 0.999,
            "epsilon": 1e-08,
            "n_iter_no_change": 10,
        }

        self.getters = {
            "activation": self.to_str,
            "solver": self.to_str,
            "alpha": self.to_float,
            "batch_size": self.to_float_or_str,
            "learning_rate": self.to_str,
            "learning_rate_init": self.to_float,
            "power_t": self.to_float,
            "max_iter": self.to_int,
            "shuffle": self.to_bool,
            "random_state": self.to_int_or_none,
            "tol": self.to_float,
            "warm_start": self.to_bool,
            "momentum": self.to_float,
            "nesterovs_momentum": self.to_bool,
            "early_stopping": self.to_bool,
            "validation_fraction": self.to_float,
            "beta_1": self.to_float,
            "beta_2": self.to_float,
            "epsilon": self.to_float,
            "n_iter_no_change": self.to_int,
        }

        self.button_importance.Disable()

        self.run()


RF_TOOL_TIPS = {
    "n_estimators": "int\nThe number of trees in the forest.",
    "criterion": "The function to measure the quality of a split. Supported criteria are "
    "“mse” for the mean squared error, which is equal to variance reduction as"
    " feature selection criterion, and “mae” for the mean absolute error.",
    "max_depth": "int, None\nThe maximum depth of the tree. If None, then nodes are expanded until all "
    "leaves are pure or until all leaves contain less than min_samples_split samples.",
    "min_samples_split": "int, float\nThe minimum number of samples required to split an "
    "internal node:\n• If int, then consider min_samples_split as the "
    "minimum number.\n• If float, then min_samples_split is a fraction "
    "and ceil(min_samples_split * n_samples) are the minimum number "
    "of samples for each split.",
    "min_samples_leaf": "int, float\nThe minimum number of samples required to be at a leaf"
    " node. A split point at any depth will only be considered if it "
    "leaves at least min_samples_leaf training samples in each of the "
    "left and right branches. This may have the effect of smoothing "
    "the model, especially in regression.\n• If int, then consider min_"
    "samples_leaf as the minimum number.\n• If float, then min_samples_"
    "leaf is a fraction and ceil(min_samples_leaf * n_samples) are the "
    "minimum number of samples for each node.",
    "min_weight_fraction_leaf": "float\nThe minimum weighted fraction of the sum total of "
    "weights (of all the input samples) required to be at a leaf"
    " node. Samples have equal weight when sample_weight is not"
    " provided.",
    "max_features": "int, float, string, or None\nThe number of features to consider when "
    "looking for the best split:\n• If int, then consider max_features "
    "features at each split.\n• If float, then max_features is a fraction"
    " and int(max_features * n_features) features are considered at each "
    "split.\n• If “auto”, then max_features=n_features.\n• If “sqrt”, "
    "then max_features=sqrt(n_features).\n• If “log2”, then max_"
    "features=log2(n_features).\n• If None, then max_features=n_features.",
    "max_leaf_nodes": "int or None\nGrow a tree with max_leaf_nodes in best-first fashion. "
    "Best nodes are defined as relative reduction in impurity. If None "
    "then unlimited number of leaf nodes.",
    "min_impurity_decrease": "float\nA node will be split if this split induces a decrease "
    "of the impurity greater than or equal to this value.\nThe "
    "weighted impurity decrease equation is the following:\n"
    "N_t / N * (impurity - N_t_R / N_t * right_impurity\n"
    "                    - N_t_L / N_t * left_impurity)\n"
    "where N is the total number of samples, N_t is the number of "
    "samples at the current node, N_t_L is the number of samples "
    "in the left child, and N_t_R is the number of samples in the "
    "right child.\nN, N_t, N_t_R and N_t_L all refer to the "
    "weighted sum, if sample_weight is passed.",
    "bootstrap": "Whether bootstrap samples are used when building trees. If False, the "
    "whole datset is used to build each tree.",
    "oob_score": "whether to use out-of-bag samples to estimate the R^2 on unseen data.",
    "n_jobs": "int or None\nThe number of jobs to run in parallel for both fit and "
    "predict. None` means 1 unless in a joblib.parallel_backend context. -1 "
    "means using all processors.",
    "random_state": "int or None\nIf int, random_state is the seed used by the random "
    "number generator; If None, the random number generator is the "
    "RandomState instance used by np.random.",
}

RF_TOOL_TIPS_CLASSIFIER = {key: value for key, value in RF_TOOL_TIPS.items()}
RF_TOOL_TIPS_CLASSIFIER["criterion"] = (
    "The function to measure the quality of a split. Supported criteria "
    "are “gini” for the Gini impurity and “entropy” for the information gain. \n"
    "Note: this parameter is tree-specific."
)

GB_TOOL_TIPS = {
    "loss": "loss function to be optimized. ‘ls’ refers to least squares regression. "
    "‘lad’ (least absolute deviation) is a highly robust loss function solely "
    "based on order information of the input variables. ‘huber’ is a combination "
    "of the two. ‘quantile’ allows quantile regression (use alpha to specify the "
    "quantile).",
    "learning_rate": "float\nlearning rate shrinks the contribution of each tree by "
    "learning_rate. There is a trade-off between learning_rate and n_estimators.",
    "n_estimators": "int\nThe number of boosting stages to perform. Gradient boosting is "
    "fairly robust to over-fitting so a large number usually results in "
    "better performance.",
    "subsample": "float\nThe fraction of samples to be used for fitting the individual "
    "base learners. If smaller than 1.0 this results in Stochastic Gradient "
    "Boosting. subsample interacts with the parameter n_estimators. Choosing "
    "subsample < 1.0 leads to a reduction of variance and an increase in bias.",
    "criterion": "The function to measure the quality of a split. Supported criteria are"
    " “friedman_mse” for the mean squared error with improvement score by "
    "Friedman, “mse” for mean squared error, and “mae” for the mean absolute"
    " error. The default value of “friedman_mse” is generally the best as it "
    "can provide a better approximation in some cases.",
    "max_depth": "int\nmaximum depth of the individual regression estimators. The maximum "
    "depth limits the number of nodes in the tree. Tune this parameter for "
    "best performance; the best value depends on the interaction of the input"
    " variables.",
    "min_samples_split": "int, float\nThe minimum number of samples required to split an "
    "internal node:\n• If int, then consider min_samples_split as "
    "the minimum number.\n• If float, then min_samples_split is a "
    "fraction and ceil(min_samples_split * n_samples) are the minimum"
    " number of samples for each split.",
    "min_samples_leaf": "int, float\nThe minimum number of samples required to be at a "
    "leaf node. A split point at any depth will only be considered if "
    "it leaves at least min_samples_leaf training samples in each of "
    "the left and right branches. This may have the effect of "
    "smoothing the model, especially in regression.\n• If int, then "
    "consider min_samples_leaf as the minimum number.\n• If float, "
    "then min_samples_leaf is a fraction and ceil(min_samples_leaf * "
    "n_samples) are the minimum number of samples for each node.",
    "min_weight_fraction_leaf": "float\nThe minimum weighted fraction of the sum total of "
    "weights (of all the input samples) required to be at a "
    "leaf node. Samples have equal weight when sample_weight is"
    " not provided.",
    "max_features": "int, float, string, or None\nThe number of features to consider when"
    " looking for the best split:\n• If int, then consider max_features"
    " features at each split.\n• If float, then max_features is a "
    "fraction and int(max_features * n_features) features are considered "
    "at each split.\n• If “auto”, then max_features=n_features.\n• If "
    "“sqrt”, then max_features=sqrt(n_features).\n• If “log2”, then max"
    "_features=log2(n_features).\n• If None, then max_features=n_features."
    "\nChoosing max_features < n_features leads to a reduction of variance"
    " and an increase in bias.",
    "alpha": "float\nThe alpha-quantile of the huber loss function and the quantile loss "
    "function. Only if loss='huber' or loss='quantile'.",
    "max_leaf_nodes": "int or None\nGrow a tree with max_leaf_nodes in best-first fashion. "
    "Best nodes are defined as relative reduction in impurity. If None "
    "then unlimited number of leaf nodes.",
    "min_impurity_decrease": "float\nA node will be split if this split induces a decrease "
    "of the impurity greater than or equal to this value.\nThe "
    "weighted impurity decrease equation is the following:\n"
    "N_t / N * (impurity - N_t_R / N_t * right_impurity\n"
    "                    - N_t_L / N_t * left_impurity)\nwhere N is"
    " the total number of samples, N_t is the number of samples at"
    " the current node, N_t_L is the number of samples in the left"
    " child, and N_t_R is the number of samples in the right child."
    "\nN, N_t, N_t_R and N_t_L all refer to the weighted sum, if "
    "sample_weight is passed.",
    "init": "If ‘zero’, the initial raw predictions are set to zero. By default a "
    "DummyEstimator is used, predicting either the average target value "
    "(for loss=’ls’), or a quantile for the other losses.",
    "random_state": "int or None\nIf int, random_state is the seed used by the random number"
    " generator; If None, the random number generator is the RandomState "
    "instance used by np.random.",
    "presort": "Whether to presort the data to speed up the finding of best splits in "
    "fitting. Auto mode by default will use presorting on dense data and default "
    "to normal sorting on sparse data. Setting presort to true on sparse data "
    "will raise an error.",
    "validation_fraction": "float\nThe proportion of training data to set aside as "
    "validation set for early stopping. Must be between 0 and 1. "
    "Only used if n_iter_no_change is set to an integer.",
    "n_iter_no_change": "int or None\nn_iter_no_change is used to decide if early stopping "
    "will be used to terminate training when validation score is not "
    "improving. By default it is set to None to disable early stopping."
    " If set to a number, it will set aside validation_fraction size of"
    " the training data as validation and terminate training when "
    "validation score is not improving in all of the previous "
    "n_iter_no_change numbers of iterations.",
    "tol": "float\nTolerance for the early stopping. When the loss is not improving by at "
    "least tol for n_iter_no_change iterations (if set to a number), the training "
    "stops.",
}

DT_TOOL_TIPS = {
    "criterion": "The function to measure the quality of a split. Supported criteria are "
    "“mse” for the mean squared error, which is equal to variance reduction "
    "as feature selection criterion and minimizes the L2 loss using the mean "
    "of each terminal node, “friedman_mse”, which uses mean squared error "
    "with Friedman’s improvement score for potential splits, and “mae” for "
    "the mean absolute error, which minimizes the L1 loss using the median of"
    " each terminal node.",
    "splitter": "The strategy used to choose the split at each node. Supported strategies "
    "are “best” to choose the best split and “random” to choose the best "
    "random split.",
    "max_depth": "int, None\nThe maximum depth of the tree. If None, then nodes are expanded"
    " until all leaves are pure or until all leaves contain less than "
    "min_samples_split samples.",
    "min_samples_split": "int, float\nThe minimum number of samples required to split an "
    "internal node:\n• If int, then consider min_samples_split as the "
    "minimum number.\n• If float, then min_samples_split is a fraction"
    " and ceil(min_samples_split * n_samples) are the minimum number "
    "of samples for each split.",
    "min_samples_leaf": "int, float\nThe minimum number of samples required to be at a leaf"
    " node. A split point at any depth will only be considered if it "
    "leaves at least min_samples_leaf training samples in each of the "
    "left and right branches. This may have the effect of smoothing the"
    " model, especially in regression.\n• If int, then consider "
    "min_samples_leaf as the minimum number.\n• If float, then "
    "min_samples_leaf is a fraction and ceil(min_samples_leaf * "
    "n_samples) are the minimum number of samples for each node.",
    "min_weight_fraction_leaf": "float\nThe minimum weighted fraction of the sum total of "
    "weights (of all the input samples) required to be at a leaf"
    " node. Samples have equal weight when sample_weight is not"
    " provided.",
    "max_features": "int, float, string, or None\nThe number of features to consider when"
    " looking for the best split:\n• If int, then consider max_features "
    "features at each split.\n• If float, then max_features is a fraction"
    " and int(max_features * n_features) features are considered at each "
    "split.\n• If “auto”, then max_features=n_features.\n• If “sqrt”, "
    "then max_features=sqrt(n_features).\n• If “log2”, "
    "then max_features=log2(n_features).\n• If None, then "
    "max_features=n_features.",
    "max_leaf_nodes": "int or None\nGrow a tree with max_leaf_nodes in best-first fashion. "
    "Best nodes are defined as relative reduction in impurity. If None "
    "then unlimited number of leaf nodes.",
    "min_impurity_decrease": "float\nA node will be split if this split induces a decrease "
    "of the impurity greater than or equal to this value.\nThe "
    "weighted impurity decrease equation is the following:\n"
    "N_t / N * (impurity - N_t_R / N_t * right_impurity\n"
    "                    - N_t_L / N_t * left_impurity)\nwhere N is"
    " the total number of samples, N_t is the number of samples at"
    " the current node, N_t_L is the number of samples in the left"
    " child, and N_t_R is the number of samples in the right child."
    "\nN, N_t, N_t_R and N_t_L all refer to the weighted sum, if "
    "sample_weight is passed.",
    "random_state": "int or None\nIf int, random_state is the seed used by the random number"
    " generator; If None, the random number generator is the RandomState "
    "instance used by np.random.",
    "presort": "Whether to presort the data to speed up the finding of best splits in "
    "fitting. For the default settings of a decision tree on large datasets, "
    "setting this to true may slow down the training process. When using either "
    "a smaller dataset or a restricted depth, this may speed up the training.",
}

SVR_TOOL_TIPS = {
    "kernel": "string\n"
    "Specifies the kernel type to be used in the algorithm. It must be one of ‘linear’, ‘poly’, "
    "‘rbf’, ‘sigmoid’, ‘precomputed’ or a callable. If none is given, ‘rbf’ will be used. If "
    "a callable is given it is used to precompute the kernel matrix.",
    "degree": "int\n"
    "Degree of the polynomial kernel function (‘poly’). Ignored by all other kernels.",
    "gamma": "float\n"
    "Kernel coefficient for ‘rbf’, ‘poly’ and ‘sigmoid’.\n\n"
    "Current default is ‘auto’ which uses 1 / n_features, if gamma='scale' is passed then it "
    "uses 1 / (n_features * X.var()) as value of gamma. The current default of gamma, ‘auto’, "
    "will change to ‘scale’ in version 0.22. ‘auto_deprecated’, a deprecated version of ‘auto’ "
    "is used as a default indicating that no explicit value of gamma was passed.",
    "coef0": "float\n"
    "Independent term in kernel function. It is only significant in ‘poly’ and ‘sigmoid’",
    "tol": "float\n" "Tolerance for stopping criterion.",
    "C": "float\n" "Penalty parameter C of the error term.",
    "epsilon": "float\n"
    "Epsilon in the epsilon-SVR model. It specifies the epsilon-tube within which no penalty is"
    " associated in the training loss function with points predicted within a distance epsilon"
    " from the actual value.",
    "shrinking": "boolean\n" "Whether to use the shrinking heuristic",
    "cache_size": "float\n" "Specify the size of the kernel cache (in MB)",
    "max_iter": "int\n"
    "Hard limit on iterations within solver, or -1 for no limit.",
}

DATA_SPLIT_TOOL_TIPS = {
    "test_size": "float, int, or None\n"
    "If float, should be between 0.0 and 1.0 and represent the proportion of the "
    "dataset to include in the test split. If int, represents the absolute number of "
    "test samples. If None, the value is set to the complement of the train size. If "
    "train_size is also None, it will be set to 0.25.",
    "train_size": "float, int, or None\n"
    "If float, should be between 0.0 and 1.0 and represent the proportion of the "
    "dataset to include in the train split. If int, represents the absolute number "
    "of train samples. If None, the value is automatically set to the complement of "
    "the test size.",
    "random_state": "int or None\n"
    "If int, random_state is the seed used by the random number generator; If None,"
    " the random number generator is the RandomState instance used by np.random.",
    "shuffle": "boolean\n"
    "Whether or not to shuffle the data before splitting. If shuffle=False then stratify"
    " must be None.",
}


MLP_TOOL_TIPS = {
    "hidden_layer_sizes": "tuple, length = n_layers - 2\n",
    "activation": "string\n"
    "Activation function for the hidden layer.\n"
    "identity: no-op activation, useful to implement linear bottleneck, returns f(x) = x.\n"
    "logistic: the logistic sigmoid function, returns f(x) = 1 / (1 + exp(-x)).\n"
    "tanh: the hyperbolic tan function, returns f(x) = tanh(x).\n"
    "relu: the rectified linear unit function, returns f(x) = max(0, x)",
    "solver": "string\nThe solver for weight optimization."
    "lbfgs is an optimizer in the family of quasi-Newton methods.\n"
    "sgd refers to stochastic gradient descent.\n"
    "adam refers to a stochastic gradient-based optimizer proposed by Kingma, Diederik, "
    "and Jimmy Ba\n\n"
    "Note: The default solver ‘adam’ works pretty well on relatively large datasets "
    "(with thousands of training samples or more) in terms of both training time and validation "
    "score. For small datasets, however, ‘lbfgs’ can converge faster and perform better.",
    "alpha": "float\n" "L2 penalty (regularization term) parameter.",
    "batch_size": "int or string\n"
    "Size of minibatches for stochastic optimizers. If the solver is ‘lbfgs’, the "
    "classifier will not use minibatch. When set to “auto”, batch_size=min(200, n_samples)",
    "learning_rate": "string\n"
    "Learning rate schedule for weight updates.\n"
    "constant is a constant learning rate given by ‘learning_rate_init’.\n"
    "invscaling gradually decreases the learning rate at each time step ‘t’ "
    "using an inverse scaling exponent of ‘power_t’. effective_"
    "learning_rate = learning_rate_init / pow(t, power_t)\n"
    "adaptive keeps the learning rate constant to ‘learning_rate_init’ as long as "
    "training loss keeps decreasing. Each time two consecutive epochs fail to decrease "
    "training loss by at least tol, or fail to increase validation score by at least "
    "tol if ‘early_stopping’ is on, the current learning rate is divided by 5.",
    "learning_rate_init": "double\n"
    "The initial learning rate used. It controls the step-size in updating the "
    "weights. Only used when solver=’sgd’ or ‘adam’.",
    "power_t": "double\n"
    "The exponent for inverse scaling learning rate. It is used in updating effective "
    "learning rate when the learning_rate is set to ‘invscaling’. Only used when solver=’sgd’.",
    "max_iter": "int\n"
    "Maximum number of iterations. The solver iterates until convergence (determined by "
    "‘tol’) or this number of iterations. For stochastic solvers (‘sgd’, ‘adam’), note that "
    "this determines the number of epochs (how many times each data point will be used), not "
    "the number of gradient steps.",
    "shuffle": "bool\n"
    "Whether to shuffle samples in each iteration. Only used when solver=’sgd’ or ‘adam’.",
    "random_state": "int or None\n"
    "If int, random_state is the seed used by the random number generator; "
    "If None, the random number generator is the RandomState instance used by np.random.",
    "tol": "float\n"
    "Tolerance for the optimization. When the loss or score is not improving by at least tol for "
    "n_iter_no_change consecutive iterations, unless learning_rate is set to ‘adaptive’, "
    "convergence is considered to be reached and training stops.",
    "warm_start": "bool\n"
    "When set to True, reuse the solution of the previous call to fit as initialization, "
    "otherwise, just erase the previous solution.",
    "momentum": "float\n"
    "Momentum for gradient descent update. Should be between 0 and 1. "
    "Only used when solver=’sgd’.",
    "nesterovs_momentum": "bool\n"
    "Whether to use Nesterov’s momentum. Only used when solver=’sgd’ "
    "and momentum > 0.",
    "early_stopping": "bool\n"
    "Whether to use early stopping to terminate training when validation score is not "
    "improving. If set to true, it will automatically set aside 10% of training data "
    "as validation and terminate training when validation score is not improving by "
    "at least tol for n_iter_no_change consecutive epochs. The split is stratified, "
    "except in a multilabel setting. Only effective when solver=’sgd’ or ‘adam’",
    "validation_fraction": "float\n"
    "The proportion of training data to set aside as validation set for early "
    "stopping. Must be between 0 and 1. Only used if early_stopping is True",
    "beta_1": "float\n"
    "Exponential decay rate for estimates of first moment vector in adam, should be in [0, 1). "
    "Only used when solver=’adam’",
    "beta_2": "float\n"
    "Exponential decay rate for estimates of second moment vector in adam, should be in [0, 1)."
    " Only used when solver=’adam’",
    "epsilon": "float\n"
    "Value for numerical stability in adam. Only used when solver=’adam’",
    "n_iter_no_change": "int\n"
    "Maximum number of epochs to not meet tol improvement. Only effective when "
    "solver=’sgd’ or ‘adam’",
}


ALGORITHMS = {
    "Random Forest": {
        "regressor": RandomForestRegressor,
        "classifier": RandomForestClassifier,
        "tool_tips": RF_TOOL_TIPS,
        "frame": RandomForestFrame,
    },
    "Support Vector Machine": {
        "regressor": SVR,
        "classifier": SVC,
        "tool_tips": SVR_TOOL_TIPS,
        "frame": SupportVectorRegressionFrame,
    },
    "Gradient Boosting": {
        "regressor": GradientBoostingRegressor,
        "classifier": GradientBoostingClassifier,
        "tool_tips": GB_TOOL_TIPS,
        "frame": GradientBoostingFrame,
    },
    "Decision Tree": {
        "regressor": DecisionTreeRegressor,
        "classifier": DecisionTreeClassifier,
        "tool_tips": DT_TOOL_TIPS,
        "frame": DecisionTreeFrame,
    },
    "Multilayer Perceptron": {
        "regressor": MLPRegressor,
        "classifier": MLPClassifier,
        "tool_tips": MLP_TOOL_TIPS,
        "frame": MLPFrame,
    },
}


class MachineLearningPlotData:
    def __init__(self, X, y, model, do_training=True, **kwargs):
        self.model = model
        self.split_args = kwargs

        indices = list(range(len(y)))

        # split the data for training and testing
        split_data = train_test_split(X, indices, **kwargs)
        self.X = {"data": X, "train": split_data[0], "test": split_data[1]}
        self.indices = {
            "data": indices,
            "train": split_data[2],
            "test": split_data[3],
        }
        self.y = {
            "data": y,
            "train": [y[i] for i in split_data[2]],
            "test": [y[i] for i in split_data[3]],
        }
        self.x = {
            key: [i + 1 for i in range(len(data))]
            for key, data in self.y.items()
        }

        # Train model, then calculate predictions, residuals, and mse
        if do_training:
            self.model.fit(self.X["train"], self.y["train"])
        self.predictions = {
            key: self.get_prediction(key) for key in self.y.keys()
        }
        self.residuals = {key: self.get_residual(key) for key in self.y.keys()}
        self.mse = {key: self.get_mse(key) for key in self.y.keys()}
        self.accuracy = {key: self.get_accuracy(key) for key in self.y.keys()}

    def get_prediction(self, key):
        return self.model.predict(self.X[key])

    def get_mse(self, key):
        return np.mean(
            np.square(np.subtract(self.predictions[key], self.y[key]))
        )

    def get_residual(self, key):
        return np.subtract(self.y[key], self.model.predict(self.X[key]))

    def get_accuracy(self, key):
        """Only applicable for classifiers"""
        return np.count_nonzero(
            np.subtract(self.predictions[key], self.y[key]) == 0
        ) / len(self.y[key])

    @property
    def feature_importances(self):
        if hasattr(self.model, "feature_importances_"):
            return self.model.feature_importances_
        return None


class FeatureImportanceFrame(wx.Frame):
    def __init__(
        self,
        options,
        x_variables,
        feature_importances,
        frame_title,
        plot_title,
    ):
        wx.Frame.__init__(self, None)

        self.title = frame_title

        self.plot = PlotFeatureImportance(
            self, options, x_variables, feature_importances, plot_title
        )

        self.set_properties()
        self.do_layout()

        self.Bind(wx.EVT_SIZE, self.on_resize)

    def set_properties(self):
        self.SetTitle(self.title)
        self.SetMinSize(get_window_size(0.35, 0.8))

    def do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.HORIZONTAL)
        sizer_wrapper.Add(self.plot.layout, 1, wx.EXPAND, 0)

        set_msw_background_color(
            self
        )  # If windows, change the background color
        set_frame_icon(self)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def redraw_plot(self):
        self.plot.redraw_plot()

    def on_resize(self, *evt):
        try:
            self.Refresh()
            self.Layout()
            wx.CallAfter(self.redraw_plot)
        except RuntimeError:
            pass


class MachineLearningModelViewer:
    def __init__(self, parent, group_data, group, options, mvr=None):
        self.parent = parent
        self.group_data = group_data
        self.group = group
        self.stats_data = group_data[group]["stats_data"]
        self.options = options

        self.file_path = self.file_select_dlg()

        if self.file_path:

            self.__load_ml_file()
            try:
                if self.is_valid:
                    self.__set_X_and_y_data()

                    self.mvr = mvr
                    self.multi_var_pred = (
                        None if mvr is None else self.mvr.predictions
                    )

                    data_keys = [
                        "X",
                        "y",
                        "x_variables",
                        "y_variable",
                        "multi_var_pred",
                        "options",
                        "mrn",
                        "study_date",
                        "uid",
                    ]
                    data = {key: getattr(self, key) for key in data_keys}
                    frame = ALGORITHMS[self.title]["frame"]
                    self.ml_frame = frame(
                        parent, data, include_test_data=False
                    )
                    self.__load_model()

                    set_msw_background_color(self.ml_frame)
                    set_frame_icon(self.ml_frame)

                    self.ml_frame.run()
                else:
                    if self.stats_data is None:
                        msg = "No data has been queried for Group %s." % group
                    elif not self.is_ml:
                        msg = "Selected file is not a valid machine learning model save file."
                    elif not self.stats_data_has_y:
                        msg = (
                            "The model's dependent variable is not found in your queried data:\n%s"
                            % self.y_variable
                        )
                    elif self.missing_x_variables:
                        msg = (
                            "Your queried data is missing the following independent variables:\n%s"
                            % ", ".join(self.missing_x_variables)
                        )
                    else:
                        msg = "Unknown error."

                    wx.MessageBox(
                        msg,
                        "Model Loading Error",
                        wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING,
                    )
            except Exception as e:
                msg = str(e)
                wx.MessageBox(
                    msg,
                    "Model Loading Error",
                    wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING,
                )

    def file_select_dlg(self):
        with wx.FileDialog(
            self.parent,
            "Load a machine learning model",
            wildcard="*.ml",
            style=wx.FD_FILE_MUST_EXIST | wx.FD_OPEN,
        ) as dlg:
            dlg.SetDirectory(MODELS_DIR)
            if dlg.ShowModal() == wx.ID_OK:
                return dlg.GetPath()

    def __load_ml_file(self):
        self.loaded_data = load_object_from_file(self.file_path)

        self.y_variable = self.loaded_data["y_variable"]
        self.tool_tips = self.loaded_data["tool_tips"]
        self.x_variables = self.loaded_data["x_variables"]
        self.title = self.loaded_data["title"]
        self.input_parameters = self.loaded_data["input_parameters"]
        self.data_split = self.loaded_data["data_split"]
        self.version = self.loaded_data["version"]

        # As of v0.8.9, model -> sklearn_predictor, regression -> model
        if "sklearn_predictor" in self.loaded_data.keys():
            self.model = self.loaded_data["model"]
            self.sklearn_predictor = self.loaded_data["sklearn_predictor"]
        else:
            self.model = self.loaded_data["regression"]
            self.sklearn_predictor = self.loaded_data["model"]

    def __load_model(self):
        self.ml_frame.model = self.model
        self.ml_frame.set_input_parameters(self.input_parameters)
        self.ml_frame.set_data_split_parameters(self.data_split)
        self.__disable_input()

    def __disable_input(self):
        for variable, input_obj in self.ml_frame.input.items():
            input_obj.Disable()
        for variable, input_obj in self.ml_frame.data_split_input.items():
            input_obj.Disable()
        self.ml_frame.button_calculate.Disable()
        self.ml_frame.button_save_model.Disable()

    def __set_X_and_y_data(self):
        (
            self.X,
            self.y,
            self.mrn,
            self.uid,
            self.study_date,
        ) = self.stats_data.get_X_and_y(
            self.y_variable, self.x_variables, include_patient_info=True
        )

    @property
    def is_ml(self):
        return "title" in list(self.loaded_data) and self.loaded_data[
            "title"
        ] in list(ALGORITHMS)

    @property
    def is_valid(self):
        return (
            self.stats_data is not None
            and self.is_ml
            and not self.missing_x_variables
            and self.stats_data_has_y
        )

    @property
    def missing_x_variables(self):
        return [
            x for x in self.x_variables if x not in list(self.stats_data.data)
        ]

    @property
    def stats_data_has_y(self):
        return self.y_variable in list(self.stats_data.data)

    def apply_plot_options(self):
        self.ml_frame.plot.apply_options()
        self.ml_frame.redraw_plot()


class MachineLearningSetupDlg(wx.Dialog):
    def __init__(self, stats_data, options):
        wx.Dialog.__init__(self, None, title="Machine Learning")

        self.stats_data = stats_data
        self.options = options

        self.combo_box_type = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=["Regression", "Classification"],
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.combo_box_alg = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=sorted(list(ALGORITHMS.keys())),
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.combo_box_nan = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=["Ignore Study", "Ignore Feature"],
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.combo_box_y = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=stats_data.variables,
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )

        self.list_ctrl_features = wx.ListCtrl(
            self, wx.ID_ANY, style=wx.LC_NO_HEADER | wx.LC_REPORT
        )

        bmp = wx.ArtProvider.GetBitmap(wx.ART_INFORMATION, size=(16, 16))
        self.button_info = wx.BitmapButton(self, id=wx.ID_ANY, bitmap=bmp)

        self.button_select_all = wx.Button(self, wx.ID_ANY, "Select All")
        self.button_deselect_all = wx.Button(self, wx.ID_ANY, "Deselect All")
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")

        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()

        self.__do_bind()
        self.__do_layout()

    def __set_properties(self):
        self.list_ctrl_features.AppendColumn(
            "Features", format=wx.LIST_FORMAT_LEFT, width=400
        )

        for choice in self.stats_data.variables:
            self.list_ctrl_features.InsertItem(50000, choice)

        self.combo_box_type.SetValue("Regression")
        self.combo_box_alg.SetValue(sorted(list(ALGORITHMS.keys()))[0])
        self.combo_box_nan.SetValue("Ignore Study")
        self.combo_box_y.SetValue(self.stats_data.variables[0])

    def __do_bind(self):
        self.Bind(
            wx.EVT_BUTTON, self.select_all, id=self.button_select_all.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.deselect_all,
            id=self.button_deselect_all.GetId(),
        )
        self.Bind(wx.EVT_BUTTON, self.on_info, id=self.button_info.GetId())

    def __do_layout(self):

        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_input_wrapper = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.HORIZONTAL
        )
        sizer_input = wx.BoxSizer(wx.VERTICAL)
        sizer_type = wx.BoxSizer(wx.VERTICAL)
        sizer_type_sub_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer_alg = wx.BoxSizer(wx.VERTICAL)
        sizer_nan_policy = wx.BoxSizer(wx.VERTICAL)
        sizer_y_var = wx.BoxSizer(wx.VERTICAL)
        sizer_features = wx.BoxSizer(wx.VERTICAL)
        sizer_select_all = wx.BoxSizer(wx.HORIZONTAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)

        label_type = wx.StaticText(self, wx.ID_ANY, "ML Type:")
        label_alg = wx.StaticText(self, wx.ID_ANY, "Algorithm:")
        label_nan_policy = wx.StaticText(self, wx.ID_ANY, "NaN Policy:")
        label_y = wx.StaticText(self, wx.ID_ANY, "Dependent Variable:")
        label_features = wx.StaticText(self, wx.ID_ANY, "Features:")

        sizer_type.Add(label_type, 0, 0, 0)
        sizer_type_sub_sizer.Add(self.combo_box_type, 1, wx.EXPAND, 0)
        sizer_type_sub_sizer.Add(self.button_info, 0, wx.EXPAND | wx.LEFT, 5)
        sizer_type.Add(sizer_type_sub_sizer, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_type, 0, wx.ALL | wx.EXPAND, 5)

        sizer_alg.Add(label_alg, 0, 0, 0)
        sizer_alg.Add(self.combo_box_alg, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_alg, 0, wx.ALL | wx.EXPAND, 5)

        sizer_nan_policy.Add(label_nan_policy, 0, 0, 0)
        sizer_nan_policy.Add(self.combo_box_nan, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_nan_policy, 0, wx.ALL | wx.EXPAND, 5)

        sizer_y_var.Add(label_y, 0, 0, 0)
        sizer_y_var.Add(self.combo_box_y, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_y_var, 0, wx.ALL | wx.EXPAND, 5)

        sizer_features.Add(label_features, 0, wx.BOTTOM | wx.EXPAND, 5)
        sizer_features.Add(self.list_ctrl_features, 1, wx.EXPAND, 0)
        sizer_select_all.Add(self.button_select_all, 0, wx.ALL, 5)
        sizer_select_all.Add(self.button_deselect_all, 0, wx.ALL, 5)
        sizer_features.Add(sizer_select_all, 0, wx.ALIGN_CENTER_HORIZONTAL, 0)

        sizer_input.Add(
            sizer_features, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5
        )

        sizer_input_wrapper.Add(sizer_input, 1, wx.EXPAND, 0)
        sizer_main.Add(sizer_input_wrapper, 1, wx.EXPAND | wx.RIGHT, 5)

        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_main.Add(sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.RIGHT, 5)

        sizer_wrapper.Add(sizer_main, 1, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(sizer_wrapper)
        sizer_wrapper.Fit(self)

        self.Layout()
        self.Center()

    def select_all(self, evt):
        self.apply_global_selection()

    def deselect_all(self, evt):
        self.apply_global_selection(on=0)

    def apply_global_selection(self, on=1):
        for i in range(len(self.stats_data.variables)):
            self.list_ctrl_features.Select(i, on=on)

    @property
    def selected_indices(self):
        if len(get_selected_listctrl_items(self.list_ctrl_features)) == 0:
            self.apply_global_selection()
        return get_selected_listctrl_items(self.list_ctrl_features)

    @property
    def selected_features(self):
        features = [
            self.list_ctrl_features.GetItem(i, 0).GetText()
            for i in self.selected_indices
        ]
        ignore_me = [self.y]
        if self.ignore_feature_if_nan:
            ignore_me.extend(self.stats_data.vars_with_nan_values)
        return sorted(list(set(features) - set(ignore_me)))

    @property
    def alg_type(self):
        return ["classifier", "regressor"][
            self.combo_box_type.GetValue() == "Regression"
        ]

    @property
    def ml_alg(self):
        return self.combo_box_alg.GetValue()

    @property
    def y(self):
        return self.combo_box_y.GetValue()

    @property
    def ignore_feature_if_nan(self):
        return self.combo_box_nan.GetValue() == "Ignore Feature"

    @property
    def ml_input_data(self):
        X, y, mrn, uid, dates = self.stats_data.get_X_and_y(
            self.y, self.selected_features, include_patient_info=True
        )

        return {
            "X": X,
            "y": y,
            "x_variables": self.selected_features,
            "y_variable": self.y,
            "options": self.options,
            "mrn": mrn,
            "study_date": dates,
            "uid": uid,
        }

    def on_info(self, *evt):
        msg = (
            "You can add new features by going to Data -> Show Stats Data in the menu bar. "
            "Right-click a column header to add a new column. You can copy/paste data to/from MS Excel.\n\n"
            "This is how you can add categorical data for a classifier (e.g., toxicity grade), however, "
            "data currently MUST be numeric. You'll need to use integers to represent your categories."
        )
        wx.MessageBox(
            msg,
            "Data Editing Tip",
            wx.OK | wx.OK_DEFAULT | wx.ICON_INFORMATION,
        )



# https://stackoverflow.com/questions/15128225/python-script-to-read-and-write-a-path-to-registry
# Accessed on May 31, 2019


import winreg
import sys
from os.path import basename


def set_reg(name, reg_path, value):
    try:
        winreg.CreateKey(winreg.HKEY_CURRENT_USER, reg_path)
        registry_key = winreg.OpenKey(
            winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_WRITE
        )
        winreg.SetValueEx(registry_key, name, 0, winreg.REG_DWORD, value)
        winreg.CloseKey(registry_key)
        return True
    except WindowsError:
        return False


def get_reg(name, reg_path):
    try:
        registry_key = winreg.OpenKey(
            winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_READ
        )
        value, regtype = winreg.QueryValueEx(registry_key, name)
        winreg.CloseKey(registry_key)
        return value
    except WindowsError:
        return None


def set_ie_emulation_level(value=11001):
    # See this site for information on which values to use
    # https://docs.microsoft.com/en-us/previous-versions/windows/internet-explorer/ie-developer/general-info/ee330730(v=vs.85)#browser_emulation
    reg_path = r"Software\Microsoft\Internet Explorer\Main\FeatureControl\FEATURE_BROWSER_EMULATION"
    set_reg(basename(sys.executable), reg_path, value)


def set_ie_lockdown_level(value=0):
    # Based on info from Venkatakrishnan at
    # https://stackoverflow.com/questions/44513580/powershell-how-to-allow-blocked-content-in-internet-explorer/44514051
    reg_path = r"Software\Microsoft\Internet Explorer\Main\FeatureControl\FEATURE_LOCALMACHINE_LOCKDOWN"
    set_reg(basename(sys.executable), reg_path, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# paths.py
"""
A collection of directories and paths updated with the script directory and user's home folder for the OS
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import sys
from os import environ
from os.path import join, dirname, expanduser, pathsep

SCRIPT_DIR = dirname(__file__)
PARENT_DIR = getattr(
    sys, "_MEIPASS", dirname(SCRIPT_DIR)
)  # PyInstaller compatibility
RESOURCES_DIR = join(SCRIPT_DIR, "resources")
ICONS_DIR = join(RESOURCES_DIR, "icons")
WIN_APP_ICON = join(ICONS_DIR, "dvha_new.ico")
LOGO_PATH = join(RESOURCES_DIR, "logo.png")
APPS_DIR = join(expanduser("~"), "Apps")
APP_DIR = join(APPS_DIR, "dvh_analytics")
PREF_DIR = join(APP_DIR, "preferences")
DATA_DIR = join(APP_DIR, "data")
INBOX_DIR = join(DATA_DIR, "inbox")
IMPORTED_DIR = join(DATA_DIR, "imported")
REVIEW_DIR = join(DATA_DIR, "review")
BACKUP_DIR = join(DATA_DIR, "backup")
TEMP_DIR = join(DATA_DIR, "temp")
MODELS_DIR = join(DATA_DIR, "models")
PROTOCOL_DIR = join(DATA_DIR, "protocols")
PROTOCOL_DEFAULT_DIR = join(RESOURCES_DIR, "protocols")
DIRECTORIES = {
    key[:-4]: value for key, value in locals().items() if key.endswith("_DIR")
}

OPTIONS_PATH = join(PREF_DIR, ".options")
OPTIONS_CHECKSUM_PATH = join(PREF_DIR, ".options_checksum")
SQL_CNF_PATH = join(PREF_DIR, "sql_connection.cnf")
LICENSE_PATH = join(RESOURCES_DIR, "LICENSE.txt")
CREATE_PGSQL_TABLES = join(SCRIPT_DIR, "db", "create_tables.sql")
CREATE_SQLITE_TABLES = join(SCRIPT_DIR, "db", "create_tables_sqlite.sql")
TG263_CSV = join(
    SCRIPT_DIR, "resources", "TG263_Nomenclature_Worksheet_20170815.csv"
)
PIP_LIST_PATH = join(SCRIPT_DIR, "resources", "pip_list")

ICONS = {
    "Open": "iconfinder_Open_1493293.png",
    "Close": "iconfinder_Close_1493281.png",
    "Save": "iconfinder_Save_1493294.png",
    "Print": "iconfinder_Print_1493286.png",
    "Export": "iconfinder_csv_file_database_extension_data_3876336.png",
    "Import": "iconfinder_import_4168538.png",
    "Settings": "iconfinder_Settings_1493289.png",
    "Database": "iconfinder_data_115746_black_edit.png",
    "ROI Map": "iconfinder_icon-map_211858.png",
    "ok-green": "iconfinder_ok-green_53916.png",
    "ko-red": "iconfinder_ko-red_53948.png",
    "custom_Y": "icon_custom_Y.png",
    "custom_X": "icon_custom_X.png",
    "rtplan": "iconfinder_Clipboard-Plan_379537_zoom.png",
    "rtstruct": "iconfinder_Education-Filled_7_3672892.png",
    "rtdose": "iconfinder_package-supported_24220.png",
    "other": "error.png",
    "studies": "iconfinder_User_Customers_1218712.png",
    "study": "iconfinder_Travel-Filled-07_3671983.png",
    "plan": "iconfinder_Clipboard-Plan_379537_zoom.png",
    "patient": "iconfinder_User_Yuppie_3_1218716.png",
    "Image": "iconfinder_m-52_4230522.png",
    "AI": "neural-network_clipart2769230_edit.png",
}
for key, value in ICONS.items():
    ICONS[key] = join(ICONS_DIR, value)


def set_phantom_js_path_environment():
    """Edit the PATH environment for PhantomJS (for Bokeh's .svg export)"""
    phantom_js_path = getattr(sys, "_MEIPASS", APP_DIR)
    if phantom_js_path not in environ["PATH"]:
        environ["PATH"] += pathsep + phantom_js_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.time_series.py
"""
Class for viewing and editing the roi map, and updating the database with changes
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from dvha.dialogs.export import save_data_to_file
from dvha.models.plot import PlotCorrelation
from dvha.dialogs.main import SelectFromListDialog
from dvha.tools.errors import ErrorDialog
from dvha.tools.utilities import get_window_size


class CorrelationFrame:
    """
    Object to be passed into notebook panel for the Time Series tab
    """

    def __init__(self, main_app):

        self.main_app = main_app
        self.parent = main_app.notebook_tab["Correlation"]
        self.options = main_app.options
        self.stats_data = main_app.group_data[1]["stats_data"]
        self.stats_data_2 = main_app.group_data[2]["stats_data"]

        self.selections = main_app.options.CORRELATION_MATRIX_VARS

        self.__define_gui_objects()
        self.__do_bind()
        self.__do_layout()

    def __define_gui_objects(self):
        self.button_var_select = wx.Button(
            self.parent, wx.ID_ANY, "Select Variables"
        )
        self.button_var_default = wx.Button(
            self.parent, wx.ID_ANY, "Default Variables"
        )
        self.button_export_csv = wx.Button(
            self.parent, wx.ID_ANY, "Export CSV"
        )
        self.button_save_figure = wx.Button(
            self.parent, wx.ID_ANY, "Save Figure"
        )
        self.plot = PlotCorrelation(self.parent, self.options)

    def __do_bind(self):
        self.parent.Bind(
            wx.EVT_BUTTON,
            self.on_var_select,
            id=self.button_var_select.GetId(),
        )
        self.parent.Bind(
            wx.EVT_BUTTON,
            self.on_var_default,
            id=self.button_var_default.GetId(),
        )
        self.parent.Bind(
            wx.EVT_BUTTON, self.export_csv, id=self.button_export_csv.GetId()
        )
        self.parent.Bind(
            wx.EVT_BUTTON, self.save_figure, id=self.button_save_figure.GetId()
        )

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_plot = wx.BoxSizer(wx.HORIZONTAL)

        sizer_buttons.Add(self.button_var_select, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_var_default, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_export_csv, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_save_figure, 0, wx.RIGHT, 5)

        sizer_plot.Add(self.plot.layout, 1, wx.EXPAND, 0)

        sizer_wrapper.Add(sizer_buttons, 0, wx.BOTTOM, 5)
        sizer_wrapper.Add(sizer_plot, 1, wx.EXPAND, 0)

        self.layout = sizer_wrapper

    def on_var_select(self, evt):
        categories = [
            c for c in list(self.stats_data.data) if "date" not in c.lower()
        ]
        categories.sort()
        size = get_window_size(1, 0.8)
        dlg = SelectFromListDialog(
            "Correlation Matrix",
            "Variables",
            categories,
            size=(350, size[1]),
            column_width=300,
            selections=self.selections,
        )
        res = dlg.ShowModal()

        if res == wx.ID_OK:
            self.selections = dlg.selected_values
            self.update_plot_data()
        dlg.Destroy()

    def on_var_default(self, evt):
        self.selections = self.options.CORRELATION_MATRIX_VARS
        self.update_plot_data()

    def set_data(self, group_data):
        self.stats_data = group_data[1]["stats_data"]
        self.stats_data_2 = group_data[2]["stats_data"]
        self.update_plot_data()

    def update_plot_data(self):
        try:
            self.plot.update_plot_data(
                self.stats_data,
                stats_data_2=self.stats_data_2,
                included_vars=self.selections,
            )
        except Exception as e:
            msg = (
                "Correlation calculation failed. Perhaps more data is needed?\n%s"
                % str(e)
            )
            ErrorDialog(self.parent, msg, "Correlation Error")

    # TODO: are these 2 functions purposeful?
    def clear_data(self):
        pass

    def get_csv(self, selection=None):
        pass

    def export_csv(self, evt):
        save_data_to_file(
            self.parent, "Export Correlation data to CSV", self.plot.get_csv()
        )

    def save_figure(self, *evt):
        self.plot.save_figure_dlg(self.parent, "Save Correlation Figure")
